# xgbmodel.py

from xgboost import XGBClassifier
from nkululeko.models.model import Model


class XGB_model(Model):
    """An XGBoost model"""

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.name = "xgb"
        self.is_classifier = True
        self.clf = XGBClassifier()  # set up the classifier

    def get_type(self):
        return "xgb"



# explore.py
# explore the feature sets

import argparse
import configparser
import os

from nkululeko.constants import VERSION
from nkululeko.experiment import Experiment
from nkululeko.utils.util import Util


def main(src_dir):
    parser = argparse.ArgumentParser(
        description="Call the nkululeko EXPLORE framework."
    )
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    module = "explore"
    expr.set_module(module)
    util = Util(module)
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")
    needs_feats = False
    try:
        # load the experiment
        expr.load(f"{util.get_save_name()}")
        needs_feats = True
    except FileNotFoundError:
        # first time: load the data
        expr.load_datasets()

        # split into train and test
        expr.fill_train_and_tests()
        util.debug(
            f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}"
        )

        plot_feats = eval(util.config_val("EXPL", "feature_distributions", "False"))
        tsne = eval(util.config_val("EXPL", "tsne", "False"))
        scatter = eval(util.config_val("EXPL", "scatter", "False"))
        spotlight = eval(util.config_val("EXPL", "spotlight", "False"))
        shap = eval(util.config_val("EXPL", "shap", "False"))
        model_type = util.config_val("EXPL", "model", False)
        plot_tree = eval(util.config_val("EXPL", "plot_tree", "False"))
        needs_feats = False
        if plot_feats or tsne or scatter or model_type or plot_tree or shap:
            # these investigations need features to explore
            expr.extract_feats()
            needs_feats = True
        # explore
    expr.analyse_features(needs_feats)
    expr.store_report()
    print("DONE")


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # use this if you want to state the config file path on command line



# aug_train.py
# train with augmentations
import ast
import os.path
import numpy as np
import configparser
import argparse
import nkululeko.experiment as exp
from nkululeko.utils.util import Util
from nkululeko.constants import VERSION
import nkululeko.glob_conf as glob_conf
from nkululeko.augment import doit as augment


def doit(config_file):
    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)

    # create a new experiment
    expr = exp.Experiment(config)
    module = "aug_train"
    expr.set_module(module)
    util = Util(module)
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")

    augmentings = util.config_val("AUGMENT", "augment", False)
    if not augmentings:
        util.error("no augmentation method specified")
    augmentings = ast.literal_eval(augmentings)
    augmentings = "_".join(augmentings)
    result_file = f"augmented_{augmentings}.csv"

    glob_conf.config["DATA"]["no_reuse"] = "True"
    glob_conf.config["FEATS"]["no_reuse"] = "True"
    glob_conf.config["AUGMENT"]["sample_selection"] = "train"
    glob_conf.config["AUGMENT"]["result"] = f"./{result_file}"
    tmp_config = "tmp.ini"
    with open(tmp_config, "w") as config_file:
        glob_conf.config.write(config_file)
    augment(tmp_config)
    databases = ast.literal_eval(config["DATA"]["databases"])
    aug_name = f"aug_{augmentings}"
    databases.append(aug_name)
    glob_conf.config["DATA"]["databases"] = str(databases)
    glob_conf.config["DATA"][aug_name] = f"{util.get_exp_dir()}/{result_file}"
    glob_conf.config["DATA"][f"{aug_name}.type"] = "csv"
    glob_conf.config["DATA"][f"{aug_name}.rename_speakers"] = "True"
    glob_conf.config["DATA"][f"{aug_name}.split_strategy"] = "train"
    util.set_config(glob_conf.config)
    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}")

    # extract features
    expr.extract_feats()

    # initialize a run manager
    expr.init_runmanager()

    # run the experiment
    reports, last_epochs = expr.run()
    result = expr.get_best_report(reports).result.test
    expr.store_report()
    print("DONE")
    return result, int(np.asarray(last_epochs).min())


def main(src_dir):
    parser = argparse.ArgumentParser(description="Call the nkululeko framework.")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"
    doit(config_file)


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # use this if you want to state the config file path on command line



# estimate.snr
"""
Module for estimating SNR (signal to noise ratio) from an audio signal.

This module provides a class `SNREstimator` which calculates the SNR based on
the log energy and energy thresholds of the audio signal.

"""

import argparse

import audiofile
import matplotlib.pyplot as plt
import numpy as np
from scipy.signal.windows import hamming


class SNREstimator:
    """Estimate SNR from audio signal using log energy and energy thresholds.

    Args:
        input_data (ndarray): Input audio signal
        sample_rate (int): Sampling rate of input audio signal
        window_size (int): Window size in samples
        hop_size (int): Hop size in samples

    Returns:
        object: SNREstimator object
        estimated_snr (float): Estimated SNR in dB, extracted from SNREstimator.estimate_snr()

        Usage:
        >>> input_data, sample_rate = audiofile.read('input.wav')
        >>> snr_estimator = SNREstimator(input_data, sample_rate, window_size=320, hop_size=160)
        >>> estimated_snr, log_energies, energy_threshold_low, energy_threshold_high = snr_estimator.estimate_snr()
    """

    def __init__(self, input_data, sample_rate, window_size=320, hop_size=160):
        self.audio_data = input_data
        self.sample_rate = sample_rate
        self.frame_length = window_size
        self.hop_length = hop_size

    def frame_audio(self, signal):
        num_frames = 1 + (len(signal) - self.frame_length) // self.hop_length
        frames = [
            signal[
                i * self.hop_length: (i * self.hop_length) + self.frame_length
            ]
            for i in range(num_frames)
        ]
        return frames

    def calculate_log_energy(self, frame):
        energy = np.sum(frame**2)
        return np.log(energy)

    def calculate_snr(self, energy_high, energy_low):
        return 10 * np.log10(energy_high / energy_low)

    def estimate_snr(self):
        frames = self.frame_audio(self.audio_data)
        log_energies = [
            self.calculate_log_energy(frame * hamming(self.frame_length))
            for frame in frames
        ]

        energy_threshold_low = np.percentile(
            log_energies, 25)  # First quartile
        energy_threshold_high = np.percentile(
            log_energies, 75
        )  # Third quartile

        low_energy_frames = [
            log_energy
            for log_energy in log_energies
            if log_energy <= energy_threshold_low
        ]
        high_energy_frames = [
            log_energy
            for log_energy in log_energies
            if log_energy >= energy_threshold_high
        ]

        mean_low_energy = np.mean(low_energy_frames)
        mean_high_energy = np.mean(high_energy_frames)

        estimated_snr = self.calculate_snr(
            np.exp(mean_high_energy), np.exp(mean_low_energy)
        )
        return (
            estimated_snr,
            log_energies,
            energy_threshold_low,
            energy_threshold_high,
        )

    def plot_energy(
        self, log_energies, energy_threshold_low, energy_threshold_high
    ):
        plt.figure(figsize=(10, 6))
        plt.plot(log_energies, label="Log Energy")
        plt.axhline(
            y=energy_threshold_low,
            color="r",
            linestyle="--",
            label="Low Energy Threshold (25th Percentile)",
        )
        plt.axhline(
            y=energy_threshold_high,
            color="g",
            linestyle="--",
            label="High Energy Threshold (75th Percentile)",
        )
        plt.xlabel("Frame")
        plt.ylabel("Log Energy")
        plt.title("Log Energy and Energy Thresholds")
        plt.legend()
        plt.show()


def main():
    parser = argparse.ArgumentParser(
        description="Estimate SNR from audio signal",
        usage=(
            "python3 estimate_snr.py -i <input_file> -ws <window_size> -hs"
            " <hop_size> -p"
        ),
    )
    parser.add_argument(
        "-i", "--input", required=True, help="Input audio file in WAV format"
    )
    parser.add_argument(
        "-ws",
        "--window_size",
        type=int,
        default=int(0.02 * 16000),
        help="Window size in samples (default: 320)",
    )
    parser.add_argument(
        "-hs",
        "--hop_size",
        type=int,
        default=int(0.01 * 16000),
        help="Hop size in samples (default: 160)",
    )
    parser.add_argument(
        "-p",
        "--plot",
        action="store_true",
        help="Plot log energy and energy thresholds",
    )
    args = parser.parse_args()

    signal, sr = audiofile.read(args.input)
    snr_estimator = SNREstimator(signal, sr, args.window_size, args.hop_size)
    estimated_snr, log_energies, energy_threshold_low, energy_threshold_high = (
        snr_estimator.estimate_snr()
    )

    print("Estimated SNR:", estimated_snr)

    if args.plot:
        snr_estimator.plot_energy(
            log_energies, energy_threshold_low, energy_threshold_high
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# process_database.py --> OREAU2
import argparse
from pathlib import Path

import pandas as pd
from nkululeko.utils import files

emotion_map = {
    "C": "anger",
    "D": "disgust",
    "J": "happiness",
    "N": "neutral",
    "P": "fear",
    "S": "surprise",
    "T": "sadness",
}

gender_map = {
    "f": "female",
    "m": "male"
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="OREAU2")
    parser.add_argument("--output_dir", type=str, default=".")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    paths = list(data_dir.glob("**/*.wav"))
    files = [file for file in paths]
    emotions = [emotion_map[file.stem[-2]] for file in files]
    speakers = [file.stem[:2] for file in files]
    genders = [gender_map[file.parts[-3]] for file in files]
    languages = ["french" for file in files]

    # convert to df
    df = pd.DataFrame({"file": files, "emotion": emotions, "speaker": speakers, "gender": genders, "language": languages})

    # allocate speakers <= 11 for test
    df_train = df[df["speaker"] > "11"]
    df_test = df.drop(df_train.index)

    # save to csv
    df.to_csv(output_dir / "oreau2.csv", index=False)
    df_train.to_csv(output_dir / "oreau2_train.csv", index=False)
    df_test.to_csv(output_dir / "oreau2_test.csv", index=False)

    print(f"Total: {len(df)}, Train: {len(df_train)}, Test: {len(df_test)}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# process_database.py --> JL

import argparse
import re
from pathlib import Path

import pandas as pd

REGEX = re.compile(r"^(?:fe)?male[12]_([a-z]+)_\d+[ab]_[12]$")


emotion_map = {
    "angry": "anger",
    "sad": "sadness",
    "neutral": "neutral",
    "happy": "happiness",
    "excited": "happiness",   # merge excited and happy
    "anxious": "anxiety",
    "apologetic": "apologetic",
    "assertive": "assertive",
    "concerned": "concern",
    "encouraging": "encouraging",
}

secondary_emotions = ["anxious", "apologetic",
                      "assertive", "concerned", "encouraging"]


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="JL")
    parser.add_argument("--output_dir", type=str, default=".")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    paths = list(
        data_dir.glob("Raw JL corpus (unchecked and unannotated)/JL(wav+txt)/*.wav")
    )
    # print p.stem with NoneType
    pstem = [p.stem for p in paths if REGEX.match(p.stem) is None]
    emotions = [emotion_map.get(REGEX.match(p.stem).group(1)) for p in paths]
    speakers = [p.stem[: p.stem.find("_")] for p in paths]
    genders = [speaker[:-1] for speaker in speakers]
    languages = ["english" for p in paths]
    countries = ["new zealand" for p in paths]

    # convert to df
    df = pd.DataFrame({"file": paths, "emotion": emotions, "speaker": speakers, "gender": genders, "language": languages, "country": countries})

    # allocate speaker male2 for test
    df_train = df[df["speaker"] != "male2"]
    df_test = df.drop(df_train.index)

    # save to csv
    df.to_csv(output_dir / "jl.csv", index=False)
    df_train.to_csv(output_dir / "jl_train.csv", index=False)
    df_test.to_csv(output_dir / "jl_test.csv", index=False)

    print(f"Total: {len(df)}, Train: {len(df_train)}, Test: {len(df_test)}")


if __name__ == "__main__":
    main()
    



# model_knn_reg.py

from sklearn.linear_model import LinearRegression
from nkululeko.models.model import Model


class Lin_reg_model(Model):
    """An KNN model"""

    is_classifier = False

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.name = "lin_reg"
        self.clf = LinearRegression()  # set up the classifier



""" reportitem.py
a basic report snippet.
"""

import os.path

class ReportItem:
    def __init__(self, topic, caption, contents, image=None):
        self.topic = topic
        self.caption = caption
        self.contents = contents
        self.has_image = False
        if image is not None:
            self.image = os.path.abspath(image)
            self.has_image = True

    def to_string(self):
        return (
            f"topic: {self.topic}, caption: {self.caption}, contents:"
            f" {self.contents}"
        )



# process_database.py --> EMOVO
# modified from: https://github.com/Strong-AI-Lab/emotion/blob/master/datasets/EMOVO/process.py

import argparse
from pathlib import Path

import pandas as pd

emotion_map = {
    "dis": "disgust",
    "gio": "happiness",
    "neu": "neutral",
    "pau": "fear",
    "rab": "anger",
    "sor": "surprise",
    "tri": "sadness",
}

gender_map = {
    "f": "female",
    "m": "male"
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="EMOVO")
    parser.add_argument("--output_dir", type=str, default=".")
    args = parser.parse_args()
    
    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    paths = list(data_dir.glob("**/*.wav"))
    files = [file for file in paths]
    emotion = [emotion_map[file.stem[:3]] for file in files]
    speaker = [file.stem[4:6] for file in files]
    gender = [gender_map[file.stem[4]] for file in files]
    language = ["italian" for file in files]

    # convert to df
    df = pd.DataFrame({"file": files, "emotion": emotion, "speaker": speaker, 
            "gender": gender, "language": language})

    # allocate last speaker f3 for test
    train_df = df[df["speaker"] != "f3"]
    test_df = df.drop(train_df.index)

    # save to csv
    df.to_csv(output_dir / "emovo.csv", index=False)
    train_df.to_csv(output_dir / "emovo_train.csv", index=False)
    test_df.to_csv(output_dir / "emovo_test.csv", index=False)

    print(f"Total: {len(df)}, Train: {len(train_df)}, Test: {len(test_df)}")
    
if __name__ == "__main__":
    main()



# resample.py
# change the sampling rate for audio file or INI file (train, test, all)

import argparse
import configparser
import os
import pandas as pd
import audformat
from nkululeko.augmenting.resampler import Resampler
from nkululeko.utils.util import Util

from nkululeko.constants import VERSION
from nkululeko.experiment import Experiment


def main(src_dir):
    parser = argparse.ArgumentParser(
        description="Call the nkululeko RESAMPLE framework.")
    parser.add_argument("--config", default=None,
                        help="The base configuration")
    parser.add_argument("--file", default=None,
                        help="The input audio file to resample")
    parser.add_argument("--replace", action="store_true",
                        help="Replace the original audio file")

    args = parser.parse_args()

    if args.file is None and args.config is None:
        print("ERROR: Either --file or --config argument must be provided.")
        exit()

    if args.file is not None:
        # Load the audio file into a DataFrame
        files = pd.Series([args.file])
        df_sample = pd.DataFrame(index=files)
        df_sample.index = audformat.utils.to_segmented_index(
            df_sample.index, allow_nat=False
        )

        # Resample the audio file
        util = Util("resampler", has_config=False)
        util.debug(f"Resampling audio file: {args.file}")
        rs = Resampler(df_sample, not_testing=True, replace=args.replace)
        rs.resample()
    else:
        # Existing code for handling INI file
        config_file = args.config

        # Test if the configuration file exists
        if not os.path.isfile(config_file):
            print(f"ERROR: no such file: {config_file}")
            exit()

        # Load one configuration per experiment
        config = configparser.ConfigParser()
        config.read(config_file)
        # Create a new experiment
        expr = Experiment(config)
        module = "resample"
        expr.set_module(module)
        util = Util(module)
        util.debug(
            f"running {expr.name} from config {config_file}, nkululeko version"
            f" {VERSION}"
        )

        if util.config_val("EXP", "no_warnings", False):
            import warnings
            warnings.filterwarnings("ignore")

        # Load the data
        expr.load_datasets()

        # Split into train and test
        expr.fill_train_and_tests()
        util.debug(
            f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}")

        sample_selection = util.config_val(
            "RESAMPLE", "sample_selection", "all")
        if sample_selection == "all":
            df = pd.concat([expr.df_train, expr.df_test])
        elif sample_selection == "train":
            df = expr.df_train
        elif sample_selection == "test":
            df = expr.df_test
        else:
            util.error(
                f"unknown selection specifier {sample_selection}, should be [all |"
                " train | test]"
            )
        util.debug(f"resampling {sample_selection}: {df.shape[0]} samples")
        replace = util.config_val("RESAMPLE", "replace", "False")
        rs = Resampler(df, replace=replace)
        rs.resample()


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)



# process_database.py --> PORTUGUESE
# modified from https://github.com/Strong-AI-Lab/emotion/blob/master/datasets/Portuguese/process.py


import argparse
import re
from pathlib import Path

import pandas as pd

REGEX = re.compile(r"^\d+[sp][AB]_([a-z]+)\d+$")
from sklearn.model_selection import train_test_split

emotion_map = {
    "angry": "anger",
    "disgust": "disgust",
    "fear": "fear",
    "happy": "happiness",
    "sad": "sadness",
    "neutral": "neutral",
    "surprise": "surprise",
}

sentence_map = {
    # Sentences
    "estaMesa": "Esta mesa é de madeira",
    "oRadio": "O rádio está ligado",
    "aqueleLivro": "Aquele livro é de história",
    "aTerra": "A Terra é um planeta",
    "oCao": "O cão trouxe a bola",
    "eleChega": "Ele chega amanhã",
    "estaRoupa": "Esta roupa é colorida",
    "osJardins": "Os jardins têm flores",
    "asPessoas": "As pessoas vão a concertos",
    "haArvores": "Há árvores na floresta",
    "osTigres": "Os tigres são selvagens",
    "oQuadro": "O quadro está na parede",
    "alguemFechou": "Alguém fechou as janelas",
    "osJovens": "Os jovens ouvem música",
    "oFutebol": "O futebol é um desporto",
    "elaViajou": "Ela viajou de comboio",
    # Pseudo-sentences
    "estaDepa": "Esta dêpa é de faneira",
    "oDarrio": "O dárrio está guilado",
    "aqueleJicro": "Aquele jicro é de hisbólia",
    "aPirra": "A Pirra é um flaneto",
    "oLao": "O lão droube a nóma",
    "eleChena": "Ele chena aguinhã",
    "estaSouda": "Esta souda é lacoripa",
    "osBartins": "Os bartins têm pléres",
    "asSemoas": "As semoas vão a cambêrtos",
    "haArjuques": "Há árjuques na plurisca",
    "osLagres": "Os lagres são siltávens",
    "oJuadre": "O juadre está na pafêne",
    "alguemBelhou": "Alguém belhou as jalétas",
    "osDofens": "Os dófens mavem tézica",
    "oDutebel": "O dutebel é um nesforpo",
    "elaJiavou": "Ela jiavou de lantóio",
}

sentence_type_map = {
    "s": "sentence",
    "p": "pseudo-sentence",
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='PORTUGUESE')
    parser.add_argument('--output_dir', type=str, default='.')
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    paths = list(data_dir.glob('*.wav'))

    sent_info = pd.concat(
        [
            pd.read_csv(
                data_dir / x,
                sep="\t",
                skiprows=5,
                header=0,
                index_col=0,
            )
            for x in [
                "Castro_2010_AppxB_Sents.txt",
                "Castro_2010_AppxB_Pseudosents.txt",
            ]
        ],
        ignore_index=True,
    ).set_index("Stimulus ")

    emotions = [emotion_map[REGEX.match(p.stem).group(1)] for p in paths]
    speakers = [p.stem[p.stem.find("_") - 1] for p in paths]
    intensity = sent_info.loc[[p.stem for p in paths],
                              "Intensity (1-7)"].tolist()
    sent_type = [sentence_type_map[p.stem[p.stem.find("_") - 2]] for p in paths]
    language = ["portuguese" for p in paths]

    df = pd.DataFrame({"file": paths, "emotion": emotions, "speaker": speakers,
                       "intensity": intensity, "sentence_type": sent_type,   
                       "language": language})

    # split into train and test
    df_train, df_test = train_test_split(df, test_size=0.2, random_state=42, 
                                         stratify=df['emotion'])
    
    # save to CSV
    df.to_csv(output_dir / 'portuguese.csv', index=False)
    df_train.to_csv(output_dir / 'portuguese_train.csv', index=False)
    df_test.to_csv(output_dir / 'portuguese_test.csv', index=False)

    print(f"Portuguese: {len(df)} samples, {len(df_train)} train, {len(df_test)} test")
    

if __name__ == '__main__':
    main()






"""
Code based on @jwagner.
"""

import dataclasses
import json
import os
import pickle
import typing

import datasets
import numpy as np
import pandas as pd
import torch
import transformers
from transformers.models.wav2vec2.modeling_wav2vec2 import Wav2Vec2Model
from transformers.models.wav2vec2.modeling_wav2vec2 import Wav2Vec2PreTrainedModel

import audeer
import audiofile
import audmetric

import nkululeko.glob_conf as glob_conf
from nkululeko.models.model import Model as BaseModel
from nkululeko.reporting.reporter import Reporter


class TunedModel(BaseModel):

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor taking the configuration and all dataframes."""
        super().__init__(df_train, df_test, feats_train, feats_test)
        super().set_model_type("finetuned")
        self.name = "finetuned_wav2vec2"
        self.target = glob_conf.config["DATA"]["target"]
        labels = glob_conf.labels
        self.class_num = len(labels)
        device = self.util.config_val("MODEL", "device", "cpu")
        self.batch_size = int(self.util.config_val("MODEL", "batch_size", "8"))
        if device != "cpu":
            self.util.debug(f"running on device {device}")
            os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
            os.environ["CUDA_VISIBLE_DEVICES"] = device
        self.df_train, self.df_test = df_train, df_test
        self.epoch_num = int(self.util.config_val("EXP", "epochs", 1))

        self._init_model()

    def _init_model(self):
        model_path = "facebook/wav2vec2-large-robust-ft-swbd-300h"
        self.num_layers = None
        self.sampling_rate = 16000
        self.max_duration_sec = 8.0
        self.accumulation_steps = 4
        # create dataset

        dataset = {}
        target_name = glob_conf.target
        data_sources = {
            "train": pd.DataFrame(self.df_train[target_name]),
            "dev": pd.DataFrame(self.df_test[target_name]),
        }

        for split in ["train", "dev"]:
            df = data_sources[split]
            y = df[target_name].astype("float")
            y.name = "targets"
            df = y.reset_index()
            df.start = df.start.dt.total_seconds()
            df.end = df.end.dt.total_seconds()
            ds = datasets.Dataset.from_pandas(df)
            dataset[split] = ds

        self.dataset = datasets.DatasetDict(dataset)

        # load pre-trained model
        le = glob_conf.label_encoder
        mapping = dict(zip(le.classes_, range(len(le.classes_))))
        target_mapping = {k: int(v) for k, v in mapping.items()}
        target_mapping_reverse = {value: key for key, value in target_mapping.items()}

        self.config = transformers.AutoConfig.from_pretrained(
            model_path,
            num_labels=len(target_mapping),
            label2id=target_mapping,
            id2label=target_mapping_reverse,
            finetuning_task=target_name,
        )
        if self.num_layers is not None:
            self.config.num_hidden_layers = self.num_layers
        setattr(self.config, "sampling_rate", self.sampling_rate)
        setattr(self.config, "data", self.util.get_data_name())

        vocab_dict = {}
        with open("vocab.json", "w") as vocab_file:
            json.dump(vocab_dict, vocab_file)
        tokenizer = transformers.Wav2Vec2CTCTokenizer("./vocab.json")
        tokenizer.save_pretrained(".")

        feature_extractor = transformers.Wav2Vec2FeatureExtractor(
            feature_size=1,
            sampling_rate=16000,
            padding_value=0.0,
            do_normalize=True,
            return_attention_mask=True,
        )
        self.processor = transformers.Wav2Vec2Processor(
            feature_extractor=feature_extractor,
            tokenizer=tokenizer,
        )
        assert self.processor.feature_extractor.sampling_rate == self.sampling_rate

        self.model = Model.from_pretrained(
            model_path,
            config=self.config,
        )
        self.model.freeze_feature_extractor()
        self.model.train()
        self.model_initialized = True

    def set_model_type(self, type):
        self.model_type = type

    def set_testdata(self, data_df, feats_df):
        self.df_test, self.feats_test = data_df, feats_df

    def reset_test(self, df_test, feats_test):
        self.df_test, self.feats_test = df_test, feats_test

    def set_id(self, run, epoch):
        self.run = run
        self.epoch = epoch
        dir = self.util.get_path("model_dir")
        name = f"{self.util.get_exp_name(only_train=True)}_{self.run}_{self.epoch:03d}.model"
        self.store_path = dir + name

    def data_collator(self, data):
        files = [d["file"] for d in data]
        starts = [d["start"] for d in data]
        ends = [d["end"] for d in data]
        targets = [d["targets"] for d in data]

        signals = []
        for file, start, end in zip(
            files,
            starts,
            ends,
        ):
            offset = start
            duration = end - offset
            if self.max_duration_sec is not None:
                duration = min(duration, self.max_duration_sec)
            signal, _ = audiofile.read(
                file,
                offset=offset,
                duration=duration,
            )
            signals.append(signal.squeeze())

        input_values = self.processor(
            signals,
            sampling_rate=self.sampling_rate,
            padding=True,
        )
        batch = self.processor.pad(
            input_values,
            padding=True,
            return_tensors="pt",
        )

        batch["labels"] = torch.tensor(targets)

        return batch

    def compute_metrics(self, p: transformers.EvalPrediction):

        metrics = {
            "UAR": audmetric.unweighted_average_recall,
            "ACC": audmetric.accuracy,
        }

        # truth = p.label_ids[:, 0].astype(int)
        truth = p.label_ids
        preds = p.predictions
        preds = np.argmax(preds, axis=1)
        scores = {}
        for name, metric in metrics.items():
            scores[f"{name}"] = metric(truth, preds)
        return scores

    def train(self):
        """Train the model."""
        model_root = self.util.get_path("model_dir")
        log_root = os.path.join(self.util.get_exp_dir(), "log")
        audeer.mkdir(log_root)
        self.torch_root = audeer.path(model_root, "torch")
        conf_file = os.path.join(self.torch_root, "config.json")
        if os.path.isfile(conf_file):
            self.util.debug(f"reusing finetuned model: {conf_file}")
            self.load(self.run, self.epoch_num)
            return
        targets = pd.DataFrame(self.dataset["train"]["targets"])
        counts = targets[0].value_counts().sort_index()
        train_weights = 1 / counts
        train_weights /= train_weights.sum()
        self.util.debug("train weights: {train_weights}")
        criterion = torch.nn.CrossEntropyLoss(
            weight=torch.Tensor(train_weights).to("cuda"),
        )
        # criterion = torch.nn.CrossEntropyLoss()

        class Trainer(transformers.Trainer):

            def compute_loss(
                self,
                model,
                inputs,
                return_outputs=False,
            ):

                targets = inputs.pop("labels").squeeze()
                targets = targets.type(torch.long)

                outputs = model(**inputs)
                logits = outputs[0].squeeze()

                loss = criterion(logits, targets)

                return (loss, outputs) if return_outputs else loss

        num_steps = (
            len(self.dataset["train"])
            // (self.batch_size * self.accumulation_steps)
            // 5
        )
        num_steps = max(1, num_steps)
        # print(num_steps)

        training_args = transformers.TrainingArguments(
            output_dir=model_root,
            logging_dir=log_root,
            per_device_train_batch_size=self.batch_size,
            per_device_eval_batch_size=self.batch_size,
            gradient_accumulation_steps=self.accumulation_steps,
            evaluation_strategy="steps",
            num_train_epochs=self.epoch_num,
            fp16=True,
            save_steps=num_steps,
            eval_steps=num_steps,
            logging_steps=num_steps,
            learning_rate=1e-4,
            save_total_limit=2,
            metric_for_best_model="UAR",
            greater_is_better=True,
            load_best_model_at_end=True,
            remove_unused_columns=False,
            report_to="none",
        )

        trainer = Trainer(
            model=self.model,
            data_collator=self.data_collator,
            args=training_args,
            compute_metrics=self.compute_metrics,
            train_dataset=self.dataset["train"],
            eval_dataset=self.dataset["dev"],
            tokenizer=self.processor.feature_extractor,
            callbacks=[transformers.integrations.TensorBoardCallback()],
        )
        trainer.train()
        trainer.save_model(self.torch_root)
        self.load(self.run, self.epoch)

    def get_predictions(self):
        results = []
        for (file, start, end), _ in audeer.progress_bar(
            self.df_test.iterrows(),
            total=len(self.df_test),
            desc=f"Predicting {len(self.df_test)} audiofiles",
        ):
            if end == pd.NaT:
                signal, sr = audiofile.read(file, offset=start)
            else:
                signal, sr = audiofile.read(
                    file, duration=end - start, offset=start, always_2d=True
                )
            assert sr == self.sampling_rate
            predictions = self.model.predict(signal)
            results.append(predictions.argmax())
        return results

    def predict(self):
        """Predict the whole eval feature set"""
        predictions = self.get_predictions()
        report = Reporter(
            self.df_test[self.target].to_numpy().astype(float),
            predictions,
            self.run,
            self.epoch_num,
        )
        return report

    def predict_sample(self, signal):
        """Predict one sample"""
        prediction = {}
        if self.util.exp_is_classification():
            # get the class probabilities
            predictions = self.model.predict(signal)
            # pred = self.clf.predict(features)
            for i in range(len(self.labels)):
                cat = self.labels[i]
                prediction[cat] = predictions[i]
        else:
            predictions = self.model.predict(signal)
            prediction = predictions
        return prediction

    def store(self):
        self.util.debug("stored: ")

    def load(self, run, epoch):
        self.set_id(run, epoch)
        self.model = Model.from_pretrained(
            self.torch_root,
            config=self.config,
        )
        # print(f"loaded model type {type(self.model)}")

    def load_path(self, path, run, epoch):
        self.set_id(run, epoch)
        with open(path, "rb") as handle:
            self.clf = pickle.load(handle)


@dataclasses.dataclass
class ModelOutput(transformers.file_utils.ModelOutput):

    logits_cat: torch.FloatTensor = None
    hidden_states: typing.Tuple[torch.FloatTensor] = None
    cnn_features: torch.FloatTensor = None


class ModelHead(torch.nn.Module):

    def __init__(self, config):

        super().__init__()

        self.dense = torch.nn.Linear(config.hidden_size, config.hidden_size)
        self.dropout = torch.nn.Dropout(config.final_dropout)
        self.out_proj = torch.nn.Linear(config.hidden_size, config.num_labels)

    def forward(self, features, **kwargs):

        x = features
        x = self.dropout(x)
        x = self.dense(x)
        x = torch.tanh(x)
        x = self.dropout(x)
        x = self.out_proj(x)

        return x


class Model(Wav2Vec2PreTrainedModel):

    def __init__(self, config):

        super().__init__(config)

        self.wav2vec2 = Wav2Vec2Model(config)
        self.cat = ModelHead(config)
        self.init_weights()

    def freeze_feature_extractor(self):
        self.wav2vec2.feature_extractor._freeze_parameters()

    def pooling(
        self,
        hidden_states,
        attention_mask,
    ):

        if attention_mask is None:  # For evaluation with batch_size==1
            outputs = torch.mean(hidden_states, dim=1)
        else:
            attention_mask = self._get_feature_vector_attention_mask(
                hidden_states.shape[1],
                attention_mask,
            )
            hidden_states = hidden_states * torch.reshape(
                attention_mask,
                (-1, attention_mask.shape[-1], 1),
            )
            outputs = torch.sum(hidden_states, dim=1)
            attention_sum = torch.sum(attention_mask, dim=1)
            outputs = outputs / torch.reshape(attention_sum, (-1, 1))

        return outputs

    def forward(
        self,
        input_values,
        attention_mask=None,
        labels=None,
        return_hidden=False,
    ):

        outputs = self.wav2vec2(
            input_values,
            attention_mask=attention_mask,
        )

        cnn_features = outputs.extract_features
        hidden_states_framewise = outputs.last_hidden_state
        hidden_states = self.pooling(
            hidden_states_framewise,
            attention_mask,
        )
        logits_cat = self.cat(hidden_states)

        if not self.training:
            logits_cat = torch.softmax(logits_cat, dim=1)

        if return_hidden:

            # make time last axis
            cnn_features = torch.transpose(cnn_features, 1, 2)

            return ModelOutput(
                logits_cat=logits_cat,
                hidden_states=hidden_states,
                cnn_features=cnn_features,
            )

        else:

            return ModelOutput(
                logits_cat=logits_cat,
            )

    def predict(self, signal):
        result = self(torch.from_numpy(signal))
        result = result[0].detach().numpy()[0]
        return result


class ModelWithPreProcessing(Model):

    def __init__(self, config):
        super().__init__(config)

    def forward(
        self,
        input_values,
    ):
        # Wav2Vec2FeatureExtractor.zero_mean_unit_var_norm():
        # normed_slice = (vector - vector[:length].mean()) / np.sqrt(vector[:length].var() + 1e-7)

        mean = input_values.mean()

        # var = input_values.var()
        # raises: onnxruntime.capi.onnxruntime_pybind11_state.NotImplemented: [ONNXRuntimeError] : 9 : NOT_IMPLEMENTED : Could not find an implementation for the node ReduceProd_3:ReduceProd(11)

        var = torch.square(input_values - mean).mean()
        input_values = (input_values - mean) / torch.sqrt(var + 1e-7)

        output = super().forward(
            input_values,
            return_hidden=True,
        )

        return (
            output.hidden_states,
            output.logits_cat,
            output.cnn_features,
        )



# feats_analyser.py
import ast
import pandas as pd
from sklearn.inspection import permutation_importance
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeRegressor
import matplotlib.pyplot as plt
from nkululeko.utils.util import Util
from nkululeko.utils.stats import normalize
from nkululeko.plots import Plots
import nkululeko.glob_conf as glob_conf
from nkululeko.reporting.report_item import ReportItem
from nkululeko.reporting.defines import Header


class FeatureAnalyser:
    def __init__(self, label, df_labels, df_features):
        self.util = Util("feats_analyser")
        self.target = self.util.config_val("DATA", "target", "emotion")
        self.labels = df_labels[self.target]
        # self.labels = df_labels["class_label"]
        self.df_labels = df_labels
        self.features = df_features
        self.label = label

    def _get_importance(self, model, permutation):
        model.fit(self.features, self.labels)
        if permutation:
            r = permutation_importance(
                model,
                self.features,
                self.labels,
                n_repeats=30,
                random_state=0,
            )
            importance = r["importances_mean"]
        else:
            importance = model.feature_importances_
        return importance

    def analyse_shap(self, model):
        """Shap analysis.

        Use the best model from a previous run and analyse feature importance with SHAP.
        https://m.mage.ai/how-to-interpret-and-explain-your-machine-learning-models-using-shap-values-471c2635b78e.
        """
        import shap

        name = "my_shap_values"
        if not self.util.exist_pickle(name):

            explainer = shap.Explainer(
                model.predict_shap,
                self.features,
                output_names=glob_conf.labels,
                algorithm="permutation",
                npermutations=5,
            )
            self.util.debug("computing SHAP values...")
            shap_values = explainer(self.features)
            self.util.to_pickle(shap_values, name)
        else:
            shap_values = self.util.from_pickle(name)
        plt.tight_layout()
        shap.plots.bar(shap_values)
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        exp_name = self.util.get_exp_name(only_data=True)
        format = self.util.config_val("PLOT", "format", "png")
        filename = f"_SHAP_{model.name}"
        filename = f"{fig_dir}{exp_name}{filename}.{format}"
        plt.savefig(filename)
        self.util.debug(f"plotted SHAP feature importance tp {filename}")

    def analyse(self):
        models = ast.literal_eval(self.util.config_val("EXPL", "model", "['log_reg']"))
        model_name = "_".join(models)
        max_feat_num = int(self.util.config_val("EXPL", "max_feats", "10"))
        # https://scikit-learn.org/stable/modules/permutation_importance.html
        permutation = eval(self.util.config_val("EXPL", "permutation", "False"))
        importance = None
        self.util.debug("analysing features...")
        result_importances = {}
        if self.util.exp_is_classification():
            for model_s in models:
                if permutation:
                    self.util.debug(
                        f"computing feature importance via permutation for {model_s}, might take longer..."
                    )
                if model_s == "bayes":
                    from sklearn.naive_bayes import GaussianNB

                    model = GaussianNB()
                    result_importances[model_s] = self._get_importance(
                        model, permutation
                    )
                elif model_s == "gmm":
                    from sklearn import mixture

                    n_components = int(
                        self.util.config_val("MODEL", "GMM_components", "4")
                    )
                    covariance_type = self.util.config_val(
                        "MODEL", "GMM_covariance_type", "full"
                    )
                    model = mixture.GaussianMixture(
                        n_components=n_components, covariance_type=covariance_type
                    )
                    result_importances[model_s] = self._get_importance(
                        model, permutation
                    )
                elif model_s == "knn":
                    from sklearn.neighbors import KNeighborsClassifier

                    method = self.util.config_val("MODEL", "KNN_weights", "uniform")
                    k = int(self.util.config_val("MODEL", "K_val", "5"))
                    model = KNeighborsClassifier(
                        n_neighbors=k, weights=method
                    )  # set up the classifier
                    result_importances[model_s] = self._get_importance(
                        model, permutation
                    )
                elif model_s == "log_reg":
                    model = LogisticRegression()
                    model.fit(self.features, self.labels)
                    if permutation:
                        r = permutation_importance(
                            model,
                            self.features,
                            self.labels,
                            n_repeats=30,
                            random_state=0,
                        )
                        importance = r["importances_mean"]
                    else:
                        importance = model.coef_[0]
                    result_importances[model_s] = importance
                elif model_s == "svm":
                    from sklearn.svm import SVC

                    c = float(self.util.config_val("MODEL", "C_val", "0.001"))
                    model = SVC(kernel="linear", C=c, gamma="scale")
                    result_importances[model_s] = self._get_importance(
                        model, permutation
                    )
                    plot_tree = eval(self.util.config_val("EXPL", "plot_tree", "False"))
                    if plot_tree:
                        plots = Plots()
                        plots.plot_tree(model, self.features)
                elif model_s == "tree":
                    model = DecisionTreeClassifier()
                    result_importances[model_s] = self._get_importance(
                        model, permutation
                    )
                    plot_tree = eval(self.util.config_val("EXPL", "plot_tree", "False"))
                    if plot_tree:
                        plots = Plots()
                        plots.plot_tree(model, self.features)
                elif model_s == "xgb":
                    from xgboost import XGBClassifier

                    model = XGBClassifier(enable_categorical=True, tree_method="hist")
                    self.labels = self.labels.astype("category")
                    result_importances[model_s] = self._get_importance(
                        model, permutation
                    )
                else:
                    self.util.error(f"invalid analysis method: {model}")
        else:  # regression experiment
            for model_s in models:
                if permutation:
                    self.util.debug(
                        f"computing feature importance via permutation for {model_s}, might take longer..."
                    )
                if model_s == "knn_reg":
                    from sklearn.neighbors import KNeighborsRegressor

                    method = self.util.config_val("MODEL", "KNN_weights", "uniform")
                    k = int(self.util.config_val("MODEL", "K_val", "5"))
                    model = KNeighborsRegressor(
                        n_neighbors=k, weights=method
                    )  # set up the classifier
                    result_importances[model_s] = self._get_importance(
                        model, permutation
                    )
                elif model_s == "lin_reg":
                    model = LinearRegression()
                    model.fit(self.features, self.labels)
                    if permutation:
                        r = permutation_importance(
                            model,
                            self.features,
                            self.labels,
                            n_repeats=30,
                            random_state=0,
                        )
                        importance = r["importances_mean"]
                    else:
                        importance = model.coef_
                    result_importances[model_s] = importance
                elif model_s == "tree_reg":
                    model = DecisionTreeRegressor()
                    result_importances[model_s] = self._get_importance(
                        model, permutation
                    )
                elif model_s == "xgr":
                    from xgboost import XGBClassifier

                    model = XGBRegressor()
                    result_importances[model_s] = self._get_importance(
                        model, permutation
                    )
                else:
                    self.util.error(f"invalid analysis method: {model_s}")
        df_imp = pd.DataFrame(
            {
                "feats": self.features.columns,
            }
        )
        for model_s in result_importances:
            if len(result_importances) == 1:
                df_imp[f"{model_s}_importance"] = result_importances[model_s]
            else:
                # normalize the distributions because they might be different
                self.util.debug(f"scaling importance values for {model_s}")
                importance = result_importances[model_s]
                importance = normalize(importance.reshape(-1, 1))
                df_imp[f"{model_s}_importance"] = importance

        df_imp["importance"] = df_imp.iloc[:, 1:].mean(axis=1).values
        df_imp = df_imp.sort_values(by="importance", ascending=False).iloc[
            :max_feat_num
        ]
        df_imp["importance"] = df_imp["importance"].map(
            lambda x: int(x * 1000) / 1000.0
        )
        ax = df_imp.plot(x="feats", y="importance", kind="bar")
        for p in ax.patches:
            ax.annotate(
                str(p.get_height()), (p.get_x() * 1.005, p.get_height() * 1.005)
            )
        title = (
            f"Feature importance for {self.label} samples with model(s) {model_name}"
        )
        if permutation:
            title += "\n based on feature permutation"
        ax.set(title=title)
        plt.tight_layout()
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        exp_name = self.util.get_exp_name(only_data=True)
        format = self.util.config_val("PLOT", "format", "png")
        filename = f"_EXPL_{model_name}"
        if permutation:
            filename += "_perm"
        filename = f"{fig_dir}{exp_name}{filename}.{format}"
        plt.savefig(filename)
        fig = ax.figure
        fig.clear()
        plt.close(fig)
        caption = f"Feature importance"
        if permutation:
            caption += " based on permutation of features."
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                caption,
                f"using {model_name} models",
                filename,
            )
        )

        # result file
        res_dir = self.util.get_path("res_dir")
        filename = f"_EXPL_{model_name}"
        if permutation:
            filename += "_perm"
        filename = f"{res_dir}{self.util.get_exp_name(only_data=True)}{filename}_{model_name}.txt"
        with open(filename, "w") as text_file:
            text_file.write(
                "features in order of decreasing importance according to model"
                f" {model_name}:\n" + f"{str(df_imp.feats.values)}\n"
            )

        df_imp.to_csv(filename, mode="a")

        # check if feature distributions should be plotted
        plot_feats = self.util.config_val("EXPL", "feature_distributions", False)
        if plot_feats:
            sample_selection = self.util.config_val("EXPL", "sample_selection", "all")
            for feature in df_imp.feats:
                # plot_feature(self, title, feature, label, df_labels, df_features):
                _plots = Plots()
                _plots.plot_feature(
                    sample_selection,
                    feature,
                    "class_label",
                    self.df_labels,
                    self.features,
                )



""" feats_mos.py
predict MOS (mean opinion score)

adapted from
from https://pytorch.org/audio/main/tutorials/squim_tutorial.html#sphx-glr-tutorials-squim-tutorial-py
paper: https://arxiv.org/pdf/2304.01448.pdf

needs 
pip uninstall -y torch torchvision torchaudio
pip install --pre torch torchvision torchaudio --extra-index-url https://download.pytorch.org/whl/nightly/cpu

"""

import os
import pandas as pd
from tqdm import tqdm
import torch
import torchaudio
from torchaudio.pipelines import SQUIM_SUBJECTIVE
from torchaudio.utils import download_asset
import audiofile
import nkululeko.glob_conf as glob_conf
from nkululeko.utils.util import Util
from nkululeko.feat_extract.featureset import Featureset


class MosSet(Featureset):
    """Class to predict MOS (mean opinion score)"""

    def __init__(self, name, data_df, feats_type):
        """Constructor. is_train is needed to distinguish from test/dev sets, because they use the codebook from the training"""
        super().__init__(name, data_df, feats_type)
        self.device = self.util.config_val("MODEL", "device", "cpu")
        self.model_initialized = False

    def init_model(self):
        # load model
        self.util.debug("loading MOS model...")
        self.subjective_model = SQUIM_SUBJECTIVE.get_model()
        NMR_SPEECH = download_asset("tutorial-assets/ctc-decoding/1688-142285-0007.wav")
        self.WAVEFORM_NMR, SAMPLE_RATE_NMR = torchaudio.load(NMR_SPEECH)
        self.model_initialized = True

    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path("store")
        store_format = self.util.config_val("FEATS", "store_format", "pkl")
        storage = f"{store}{self.name}.{store_format}"
        extract = self.util.config_val("FEATS", "needs_feature_extraction", False)
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            if not self.model_initialized:
                self.init_model()
            self.util.debug("predicting MOS, this might take a while...")
            emb_series = pd.Series(index=self.data_df.index, dtype=object)
            length = len(self.data_df.index)
            for idx, (file, start, end) in enumerate(
                tqdm(self.data_df.index.to_list())
            ):
                signal, sampling_rate = audiofile.read(
                    file,
                    offset=start.total_seconds(),
                    duration=(end - start).total_seconds(),
                    always_2d=True,
                )
                emb = self.get_embeddings(signal, sampling_rate, file)
                emb_series[idx] = emb
            self.df = pd.DataFrame(emb_series.values.tolist(), index=self.data_df.index)
            self.df.columns = ["mos"]
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug("reusing predicted MOS values")
            self.df = self.util.get_store(storage, store_format)
            if self.df.isnull().values.any():
                nanrows = self.df.columns[self.df.isna().any()].tolist()
                print(nanrows)
                self.util.error(
                    f"got nan: {self.df.shape} {self.df.isnull().sum().sum()}"
                )

    def get_embeddings(self, signal, sampling_rate, file):
        tmp_audio_name = "mos_audio_tmp.wav"
        try:
            audiofile.write(tmp_audio_name, signal, sampling_rate)
            WAVEFORM_SPEECH, SAMPLE_RATE_SPEECH = torchaudio.load(tmp_audio_name)
            with torch.no_grad():
                mos = self.subjective_model(WAVEFORM_SPEECH, self.WAVEFORM_NMR)
        except RuntimeError as re:
            print(str(re))
            self.util.error(f"couldn't extract file: {file}")
        return float(mos[0].numpy())

    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_embeddings(signal, sr, "no file")
        return feats



import ast
import glob
import json
import math

from confidence_intervals import evaluate_with_conf_int
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import pearsonr
from sklearn.metrics import ConfusionMatrixDisplay
from sklearn.metrics import classification_report
from sklearn.metrics import confusion_matrix
from sklearn.metrics import r2_score
from torch import is_tensor

from audmetric import accuracy
from audmetric import concordance_cc
from audmetric import mean_absolute_error
from audmetric import mean_squared_error
from audmetric import unweighted_average_recall

import nkululeko.glob_conf as glob_conf
from nkululeko.reporting.defines import Header
from nkululeko.reporting.report_item import ReportItem
from nkululeko.reporting.result import Result
from nkululeko.utils.util import Util


class Reporter:
    def __set_measure(self):
        if self.util.exp_is_classification():
            self.MEASURE = "UAR"
            self.result.measure = self.MEASURE
            self.is_classification = True
        else:
            self.is_classification = False
            self.measure = self.util.config_val("MODEL", "measure", "mse")
            if self.measure == "mse":
                self.MEASURE = "MSE"
                self.result.measure = self.MEASURE
            elif self.measure == "mae":
                self.MEASURE = "MAE"
                self.result.measure = self.MEASURE
            elif self.measure == "ccc":
                self.MEASURE = "CCC"
                self.result.measure = self.MEASURE

    def __init__(self, truths, preds, run, epoch):
        """Initialization with ground truth und predictions vector."""
        self.util = Util("reporter")
        self.format = self.util.config_val("PLOT", "format", "png")
        self.truths = np.asarray(truths)
        self.preds = np.asarray(preds)
        self.result = Result(0, 0, 0, 0, "unknown")
        self.run = run
        self.epoch = epoch
        self.__set_measure()
        self.filenameadd = ""
        self.cont_to_cat = False
        if len(self.truths) > 0 and len(self.preds) > 0:
            if self.util.exp_is_classification():
                uar, (upper, lower) = evaluate_with_conf_int(
                    self.preds,
                    unweighted_average_recall,
                    self.truths,
                    num_bootstraps=1000,
                    alpha=5,
                )
                self.result.test = uar
                self.result.set_upper_lower(upper, lower)
                self.result.loss = 1 - accuracy(self.truths, self.preds)
            else:
                # regression experiment
                if self.measure == "mse":
                    test_result, (upper, lower) = evaluate_with_conf_int(
                        self.preds,
                        mean_squared_error,
                        self.truths,
                        num_bootstraps=1000,
                        alpha=5,
                    )
                elif self.measure == "mae":
                    test_result, (upper, lower) = evaluate_with_conf_int(
                        self.preds,
                        mean_absolute_error,
                        self.truths,
                        num_bootstraps=1000,
                        alpha=5,
                    )
                elif self.measure == "ccc":
                    test_result, (upper, lower) = evaluate_with_conf_int(
                        self.preds,
                        concordance_cc,
                        self.truths,
                        num_bootstraps=1000,
                        alpha=5,
                    )

                    if math.isnan(self.result.test):
                        self.util.debug(f"Truth: {self.truths}")
                        self.util.debug(f"Predict.: {self.preds}")
                        self.util.debug("Result is NAN: setting to -1")
                        self.result.test = -1
                else:
                    self.util.error(f"unknown measure: {self.measure}")

                self.result.test = test_result
                self.result.set_upper_lower(upper, lower)
                # train and loss are being set by the model

    def set_id(self, run, epoch):
        """Make the report identifiable with run and epoch index."""
        self.run = run
        self.epoch = epoch

    def continuous_to_categorical(self):
        if self.cont_to_cat:
            return
        self.cont_to_cat = True
        bins = ast.literal_eval(glob_conf.config["DATA"]["bins"])
        self.truths = np.digitize(self.truths, bins) - 1
        self.preds = np.digitize(self.preds, bins) - 1

    def plot_confmatrix(self, plot_name, epoch):
        if not self.util.exp_is_classification():
            self.continuous_to_categorical()
        self._plot_confmat(self.truths, self.preds, plot_name, epoch)

    def plot_per_speaker(self, result_df, plot_name, function):
        """Plot a confusion matrix with the mode category per speakers.

        Args:
            result_df: a pandas dataframe with columns: preds, truths and speaker.
            plot_name: name for the figure.
            function: either mode or mean.
        """
        speakers = result_df.speaker.unique()
        pred = np.zeros(0)
        truth = np.zeros(0)
        for s in speakers:
            s_df = result_df[result_df.speaker == s]
            mode = s_df.pred.mode().iloc[-1]
            mean = s_df.pred.mean()
            if function == "mode":
                s_df.pred = mode
            elif function == "mean":
                s_df.pred = mean
            else:
                self.util.error(f"unkown function {function}")
            pred = np.append(pred, s_df.pred.values)
            truth = np.append(truth, s_df["truth"].values)
        if not (self.is_classification or self.cont_to_cat):
            bins = ast.literal_eval(glob_conf.config["DATA"]["bins"])
            truth = np.digitize(truth, bins) - 1
            pred = np.digitize(pred, bins) - 1
        self._plot_confmat(truth, pred.astype("int"), plot_name, 0)

    def _plot_confmat(self, truths, preds, plot_name, epoch):
        # print(truths)
        # print(preds)
        fig_dir = self.util.get_path("fig_dir")
        labels = glob_conf.labels
        fig = plt.figure()  # figsize=[5, 5]
        uar, (upper, lower) = evaluate_with_conf_int(
            self.preds,
            unweighted_average_recall,
            self.truths,
            num_bootstraps=1000,
            alpha=5,
        )
        acc = accuracy(truths, preds)
        cm = confusion_matrix(
            truths, preds, normalize=None
        )  # normalize must be one of {'true', 'pred', 'all', None}
        if cm.shape[0] != len(labels):
            self.util.error(
                f"mismatch between confmatrix dim ({cm.shape[0]}) and labels"
                f" length ({len(labels)}: {labels})"
            )

        try:
            disp = ConfusionMatrixDisplay(
                confusion_matrix=cm, display_labels=labels
            ).plot(cmap="Blues")
        except ValueError:
            disp = ConfusionMatrixDisplay(
                confusion_matrix=cm,
                display_labels=list(labels).remove("neutral"),
            ).plot(cmap="Blues")

        reg_res = ""
        if not self.is_classification:
            reg_res = f"{self.result.test:.3f} {self.MEASURE}"

        uar_str = str(int(uar * 1000) / 1000.0)[1:]
        acc_str = str(int(acc * 1000) / 1000.0)[1:]
        up_str = str(int(upper * 1000) / 1000.0)[1:]
        low_str = str(int(lower * 1000) / 1000.0)[1:]

        if epoch != 0:
            plt.title(
                f"Confusion Matrix, UAR: {uar_str} "
                + f"(+-{up_str}/{low_str}), {reg_res}, Epoch: {epoch}"
            )
        else:
            plt.title(
                f"Confusion Matrix, UAR: {uar_str} "
                + f"(+-{up_str}/{low_str}) {reg_res}"
            )
        img_path = f"{fig_dir}{plot_name}{self.filenameadd}.{self.format}"
        plt.savefig(img_path)
        fig.clear()
        plt.close(fig)
        plt.savefig(img_path)
        plt.close(fig)
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_RESULTS,
                self.util.get_model_description(),
                "Confusion matrix",
                img_path,
            )
        )

        res_dir = self.util.get_path("res_dir")
        rpt = (
            f"epoch: {epoch}, UAR: {uar_str}"
            + f", (+-{up_str}/{low_str}), ACC: {acc_str}"
        )
        # print(rpt)
        self.util.debug(rpt)
        file_name = f"{res_dir}{self.util.get_exp_name()}{self.filenameadd}_conf.txt"
        with open(file_name, "w") as text_file:
            text_file.write(rpt)

    def set_filename_add(self, my_string):
        self.filenameadd = f"_{my_string}"

    def print_results(self, epoch):
        """Print all evaluation values to text file."""
        res_dir = self.util.get_path("res_dir")
        file_name = f"{res_dir}{self.util.get_exp_name()}_{epoch}{self.filenameadd}.txt"
        if self.util.exp_is_classification():
            labels = glob_conf.labels
            try:
                rpt = classification_report(
                    self.truths,
                    self.preds,
                    target_names=labels,
                    output_dict=True,
                )
            except ValueError as e:
                self.util.debug(
                    "Reporter: caught a ValueError when trying to get"
                    " classification_report: " + e
                )
                rpt = self.result.to_string()
            with open(file_name, "w") as text_file:
                c_ress = list(range(len(labels)))
                for i, l in enumerate(labels):
                    c_res = rpt[l]["f1-score"]
                    c_ress[i] = float(f"{c_res:.3f}")
                self.util.debug(f"labels: {labels}")
                f1_per_class = f"result per class (F1 score): {c_ress}"
                self.util.debug(f1_per_class)
                rpt_str = f"{json.dumps(rpt)}\n{f1_per_class}"
                text_file.write(rpt_str)
                glob_conf.report.add_item(
                    ReportItem(
                        Header.HEADER_RESULTS,
                        f"Classification result {self.util.get_model_description()}",
                        rpt_str,
                    )
                )

        else:  # regression
            result = self.result.test
            r2 = r2_score(self.truths, self.preds)
            pcc = pearsonr(self.truths, self.preds)[0]
            measure = self.util.config_val("MODEL", "measure", "mse")
            with open(file_name, "w") as text_file:
                text_file.write(
                    f"{measure}: {result:.3f}, r_2: {r2:.3f}, pcc {pcc:.3f}"
                )

    def make_conf_animation(self, out_name):
        import imageio

        fig_dir = self.util.get_path("fig_dir")
        filenames = glob.glob(fig_dir + f"{self.util.get_plot_name()}*_?_???_cnf.png")
        images = []
        for filename in filenames:
            images.append(imageio.imread(filename))
        fps = self.util.config_val("PLOT", "fps", "1")
        try:
            imageio.mimsave(fig_dir + out_name, images, fps=int(fps))
        except RuntimeError as e:
            self.util.error("error writing anim gif: " + e)

    def get_result(self):
        return self.result

    def plot_epoch_progression(self, reports, out_name):
        fig_dir = self.util.get_path("fig_dir")
        results, losses, train_results, losses_eval = [], [], [], []
        for r in reports:
            results.append(r.get_result().test)
            losses.append(r.get_result().loss)
            train_results.append(r.get_result().train)
            losses_eval.append(r.get_result().loss_eval)

        # do a plot per run
        # scale the losses so they fit on the picture
        losses, results, train_results, losses_eval = (
            np.asarray(losses),
            np.asarray(results),
            np.asarray(train_results),
            np.asarray(losses_eval),
        )

        if np.all((results > 1)):
            # scale down values
            results = results / 100.0
            train_results = train_results / 100.0
        # if np.all((losses < 1)):
        # scale up values
        plt.figure(dpi=200)
        plt.plot(train_results, "green", label="train set")
        plt.plot(results, "red", label="dev set")
        plt.plot(losses, "black", label="losses")
        plt.plot(losses_eval, "grey", label="losses_eval")
        plt.xlabel("epochs")
        plt.ylabel(f"{self.MEASURE}")
        plt.legend()
        plt.savefig(f"{fig_dir}{out_name}.{self.format}")
        plt.close()



"""
feats_spectra.py

Inspired by code from Su Lei

"""

import os
import torchaudio
import torchaudio.transforms as T
import torch
from torch.utils.data import Dataset
from PIL import Image, ImageOps
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
import pathlib
import audeer

from nkululeko.feat_extract.featureset import Featureset
from nkululeko.constants import SAMPLING_RATE
import nkululeko.glob_conf as glob_conf


class Spectraloader(Featureset):
    def __init__(self, name, data_df, feat_type):
        """Constructor setting the name"""
        super().__init__(name, data_df, feat_type)
        self.sampling_rate = SAMPLING_RATE
        self.num_bands = int(self.util.config_val("FEATS", "fft_nbands", "64"))
        self.win_dur = int(self.util.config_val("FEATS", "fft_win_dur", "25"))
        self.hop_dur = int(self.util.config_val("FEATS", "fft_hop_dur", "10"))

    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path("store")
        store_format = self.util.config_val("FEATS", "store_format", "pkl")
        storage = f"{store}{self.name}.{store_format}"
        extract = self.util.config_val("FEATS", "needs_feature_extraction", False)
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            self.util.debug("extracting mel spectra, this might take a while...")
            image_store = audeer.mkdir(f"{store}{self.name}")
            images = []
            for idx, (file, start, end) in enumerate(
                tqdm(self.data_df.index.to_list())
            ):
                signal, sampling_rate = torchaudio.load(
                    file,
                    frame_offset=int(start.total_seconds() * 16000),
                    num_frames=int((end - start).total_seconds() * 16000),
                )
                assert sampling_rate == 16000, f"got {sampling_rate} instead of 16000"
                image = self._waveform2rgb(signal)
                outfile = f"{image_store}/{pathlib.Path(file).stem}_{idx}.jpg"
                image.save(outfile)
                images.append(outfile)
            self.df = pd.DataFrame(images, index=self.data_df.index)
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug("reusing extracted spectrograms")
            self.df = self.util.get_store(storage, store_format)

    def _waveform2rgb(self, waveform, target_size=(256, 256)):
        # Transform to spectrogram
        spectrogram = T.MelSpectrogram(
            sample_rate=SAMPLING_RATE,
            n_mels=self.num_bands,
            hop_length=int(self.hop_dur * SAMPLING_RATE / 1000),
            win_length=int(self.win_dur * SAMPLING_RATE / 1000),
        )(waveform)
        melspec = T.AmplitudeToDB()(spectrogram)[0].numpy()
        melspec_norm = (melspec - np.min(melspec)) / (np.max(melspec) - np.min(melspec))

        # Map normalized Mel spectrogram to viridis colormap
        cmapped = plt.get_cmap("viridis")(melspec_norm)

        # Convert this colormap representation to a format suitable for creating a PIL Image
        image_array = (cmapped[:, :, :3] * 255).astype(np.uint8)
        image = Image.fromarray(image_array, mode="RGB")
        image = ImageOps.flip(image)

        # Resize to target size
        image = image.resize(target_size, Image.Resampling.LANCZOS)
        return image



# feats_import.py

import os
import ast
import audformat
import pandas as pd
from nkululeko.utils.util import Util
from nkululeko.feat_extract.featureset import Featureset


class ImportSet(Featureset):
    """Class to import features that have been compiled elsewhere"""

    def __init__(self, name, data_df, feats_type):
        super().__init__(name, data_df, feats_type)

    def extract(self):
        """Import the features."""
        self.util.debug(f"importing features for {self.name}")
        try:
            feat_import_files = self.util.config_val("FEATS", "import_file", False)
            feat_import_files = ast.literal_eval(feat_import_files)
        except ValueError as e:
            self.util.error(
                "feature type == import needs import_file = ['file1', 'filex']"
            )
        except SyntaxError as se:
            if type(feat_import_files) == str:
                feat_import_files = [feat_import_files]
            else:
                self.util.error(f"import_file is wrong: {feat_import_files}")

        feat_df = pd.DataFrame()
        for feat_import_file in feat_import_files:
            if not os.path.isfile(feat_import_file):
                self.util.error(f"no import file: {feat_import_file}")
            df = audformat.utils.read_csv(feat_import_file)
            df = df[df.index.isin(self.data_df.index)]
            feat_df = pd.concat([feat_df, df])
        if feat_df.shape[0] == 0:
            self.util.error(f"Imported features for data set {self.name} not found!")
        # and assign to be the "official" feature set
        self.df = feat_df



# model.py
import ast
import pickle
import random

import numpy as np
import pandas as pd
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import LeaveOneGroupOut
from sklearn.model_selection import StratifiedKFold
import sklearn.utils

import nkululeko.glob_conf as glob_conf
from nkululeko.reporting.reporter import Reporter
from nkululeko.utils.util import Util


class Model:
    """Generic model class for linear (non-neural) algorithms."""

    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor taking the configuration and all dataframes."""
        self.name = "undefined"
        self.df_train, self.df_test, self.feats_train, self.feats_test = (
            df_train,
            df_test,
            feats_train,
            feats_test,
        )
        self.model_type = "classic"
        self.util = Util("model")
        self.target = self.util.config_val("DATA", "target", "emotion")
        self.run = 0
        self.epoch = 0
        self.logo = self.util.config_val("MODEL", "logo", False)
        self.xfoldx = self.util.config_val("MODEL", "k_fold_cross", False)

    def set_model_type(self, type):
        self.model_type = type

    def is_ann(self):
        if (self.model_type == "ann") or (self.model_type == "finetuned"):
            return True
        else:
            return False

    def set_testdata(self, data_df, feats_df):
        self.df_test, self.feats_test = data_df, feats_df

    def reset_test(self, df_test, feats_test):
        self.df_test, self.feats_test = df_test, feats_test

    def set_id(self, run, epoch):
        self.run = run
        self.epoch = epoch
        dir = self.util.get_path("model_dir")
        name = f"{self.util.get_exp_name(only_train=True)}_{self.run}_{self.epoch:03d}.model"
        self.store_path = dir + name

    def _x_fold_cross(self):
        # ignore train and test sets and do a x-fold-cross  evaluation
        self.util.debug(
            f"ignoring splits and doing {self.xfoldx} fold cross validation"
        )
        feats = pd.concat([self.feats_train, self.feats_test])
        annos = pd.concat([self.df_train, self.df_test])
        targets = annos[self.target]
        _skf = StratifiedKFold(n_splits=int(self.xfoldx))
        truths, preds, results = [], [], []
        g_index = 0
        # leave-one-speaker loop
        for train_index, test_index in _skf.split(
            feats,
            targets,
        ):
            train_x = feats.iloc[train_index].to_numpy()
            train_y = targets[train_index]
            self.clf.fit(train_x, train_y)
            truth_x = feats.iloc[test_index].to_numpy()
            truth_y = targets[test_index]
            predict_y = self.clf.predict(truth_x)
            report = Reporter(truth_y.astype(float), predict_y, self.run, self.epoch)
            self.util.debug(
                f"result for fold {g_index}:"
                f" {report.get_result().get_test_result()} "
            )
            results.append(float(report.get_result().test))
            truths.append(truth_y)
            preds.append(predict_y)
            g_index += 1

        # combine speaker folds
        truth = pd.concat(truths)
        truth.name = "truth"
        pred = pd.Series(
            np.concatenate(preds),
            index=truth.index,
            name="prediction",
        )
        self.truths = truth
        self.preds = pred
        results = np.asarray(results)
        self.util.debug(
            f"KFOLD: {self.xfoldx} folds: mean {results.mean():.3f}, std:"
            f" {results.std():.3f}"
        )

    def _do_logo(self):
        # ignore train and test sets and do a "leave one speaker group out"  evaluation
        logo = int(self.logo)
        feats = pd.concat([self.feats_train, self.feats_test])
        annos = pd.concat([self.df_train, self.df_test])
        targets = annos[self.target]
        _logo = LeaveOneGroupOut()
        truths, preds, results = [], [], []
        # get unique list of speakers
        speakers = annos["speaker"].unique()
        # check for folds columns
        if not "fold" in annos.columns:
            self.util.debug(f"creating random folds for {logo} groups")
            # create a random dictionary of groups
            sdict = {}
            # randomize the speaker order
            random.shuffle(speakers)
            folds = list(range(logo))
            for i, s in enumerate(speakers):
                sdict[s] = folds[i % len(folds)]
            # add this to the annotations
            annos["fold"] = annos["speaker"].apply(lambda x: str(sdict[x]))
            fold_count = self.logo
        else:
            fold_count = annos["fold"].nunique()
            self.util.debug(f"using existing folds for {fold_count} groups")
        g_index = 0
        self.util.debug(f"ignoring splits and doing LOGO with {fold_count} groups")
        # leave-one-group loop
        for train_index, test_index in _logo.split(
            feats,
            targets,
            groups=annos["fold"],
        ):
            train_x = feats.iloc[train_index].to_numpy()
            train_y = targets.iloc[train_index]
            self.clf.fit(train_x, train_y)

            truth_x = feats.iloc[test_index].to_numpy()
            truth_y = targets.iloc[test_index]
            predict_y = self.clf.predict(truth_x)
            report = Reporter(truth_y.astype(float), predict_y, self.run, self.epoch)
            result = report.get_result().get_test_result()
            self.util.debug(f"result for speaker group {g_index}: {result} ")
            results.append(float(report.get_result().test))
            truths.append(truth_y)
            preds.append(predict_y)
            g_index += 1

        # combine speaker folds
        truth = pd.concat(truths)
        truth.name = "truth"
        pred = pd.Series(
            np.concatenate(preds),
            index=truth.index,
            name="prediction",
        )
        self.truths = truth
        self.preds = pred
        results = np.asarray(results)
        self.util.debug(
            f"LOGO: {self.logo} folds: mean {results.mean():.3f}, std:"
            f" {results.std():.3f}"
        )

    def train(self):
        """Train the model"""
        # # first check if the model already has been trained
        # if os.path.isfile(self.store_path):
        #     self.load(self.run, self.epoch)
        #     self.util.debug(f'reusing model: {self.store_path}')
        #     return

        # then if leave one speaker group out validation is wanted
        if self.logo:
            self._do_logo()
            return
        # then if x fold cross validation is wanted
        if self.xfoldx:
            self._x_fold_cross()
            return

        # check for NANs in the features
        # set up the data_loaders
        if self.feats_train.isna().to_numpy().any():
            self.util.debug(
                "Model, train: replacing"
                f" {self.feats_train.isna().sum().sum()} NANs with 0"
            )
            self.feats_train = self.feats_train.fillna(0)
        # remove labels from features
        feats = self.feats_train.to_numpy()
        # compute class weights
        if self.util.config_val("MODEL", "class_weight", False):
            self.classes_weights = sklearn.utils.class_weight.compute_sample_weight(
                class_weight="balanced", y=self.df_train[self.target]
            )

        tuning_params = self.util.config_val("MODEL", "tuning_params", False)
        if tuning_params:
            # tune the model meta parameters
            tuning_params = ast.literal_eval(tuning_params)
            tuned_params = {}
            try:
                scoring = glob_conf.config["MODEL"]["scoring"]
            except KeyError:
                self.util.error("got tuning params but no scoring")
            for param in tuning_params:
                values = ast.literal_eval(glob_conf.config["MODEL"][param])
                tuned_params[param] = values
            self.util.debug(f"tuning on {tuned_params}")
            self.clf = GridSearchCV(
                self.clf, tuned_params, refit=True, verbose=3, scoring=scoring
            )
            try:
                class_weight = self.util.config_val("MODEL", "class_weight", False)
                if class_weight:
                    self.util.debug("using class weight")
                    self.clf.fit(
                        feats,
                        self.df_train[self.target],
                        sample_weight=self.classes_weights,
                    )
                else:
                    self.clf.fit(feats, self.df_train[self.target])
            except KeyError:
                self.clf.fit(feats, self.df_train[self.target])
            self.util.debug(f"winner parameters: {self.clf.best_params_}")
        else:
            class_weight = self.util.config_val("MODEL", "class_weight", False)
            if class_weight:
                self.util.debug("using class weight")
                self.clf.fit(
                    feats,
                    self.df_train[self.target],
                    sample_weight=self.classes_weights,
                )
            else:
                labels = self.df_train[self.target]
                self.clf.fit(feats, labels)

    def get_predictions(self):
        predictions = self.clf.predict(self.feats_test.to_numpy())
        return predictions

    def predict(self):
        if self.feats_test.isna().to_numpy().any():
            self.util.debug(
                "Model, test: replacing"
                f" {self.feats_test.isna().sum().sum()} NANs with 0"
            )
            self.feats_test = self.feats_test.fillna(0)
        if self.logo or self.xfoldx:
            report = Reporter(
                self.truths.astype(float), self.preds, self.run, self.epoch
            )
            return report
        """Predict the whole eval feature set"""
        predictions = self.get_predictions()
        report = Reporter(
            self.df_test[self.target].to_numpy().astype(float),
            predictions,
            self.run,
            self.epoch,
        )
        return report

    def get_type(self):
        return "generic"

    def predict_sample(self, features):
        """Predict one sample"""
        prediction = {}
        if self.util.exp_is_classification():
            # get the class probabilities
            predictions = self.clf.predict_proba(features)
            # pred = self.clf.predict(features)
            for i in range(len(self.clf.classes_)):
                cat = self.clf.classes_[i]
                prediction[cat] = predictions[0][i]
        else:
            predictions = self.clf.predict(features)
            prediction = predictions[0]
        return prediction

    def store(self):
        with open(self.store_path, "wb") as handle:
            pickle.dump(self.clf, handle)

    def load(self, run, epoch):
        self.set_id(run, epoch)
        dir = self.util.get_path("model_dir")
        name = f"{self.util.get_exp_name(only_train=True)}_{self.run}_{self.epoch:03d}.model"
        try:
            with open(dir + name, "rb") as handle:
                self.clf = pickle.load(handle)
        except FileNotFoundError as fe:
            self.util.error(
                f"Did you forget to store your models? needs: \n[MODEL]\nsave=True\n{fe}"
            )

    def load_path(self, path, run, epoch):
        self.set_id(run, epoch)
        with open(path, "rb") as handle:
            self.clf = pickle.load(handle)



# process_database --> ASVP-ESD

import argparse
import os
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

emotion_map = {
    '01': 'boredom',
    '02': 'neutral',
    '03': 'happy',
    '04': 'sad',
    '05': 'anger',
    '06': 'fear',
    '07': 'disgust',
    '08': 'surprise',
    '09': 'excited',
    '10': 'pleasure',
    '11': 'pain',
    '12': 'disapointed',
    '13': 'others',
}


vocal_map = {
    "01" : "speech",
    "02" : "non_speech",
}

language_map = {
    "00" : "chinese",
    "01" : "english",
    "02" : "french",
    "03" : "russian",
    "04" : "other1",
    "05" : "other2",
}

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=Path,
        default=Path('./ASVP-ESD-Update'))
    parser.add_argument('--output_dir', type=Path, default=Path('./'))
    args = parser.parse_args()

    # Read the database
    input_dir = args.data_dir
    output_dir = args.output_dir

    # list of all the WAV files in the database
    wav_list = list(input_dir.glob('Audio/**/*.wav'))

    # file = [p for p in wav_list if os.path.getsize(p) >= 40000]
    file = [p for p in wav_list]

    # Vocal channel (01 = speech, 02 = non speech).
    vocal = [vocal_map[p.stem.split('-')[1]] for p in file]

    # emotion
    emotion = [emotion_map[p.stem.split('-')[2]] for p in file]

    # language --> not all data have language
    # language = [language_map[p.stem.split('-')[8]] for p in wav_list]

    # save to pandas dataframe
    df = pd.DataFrame(data={'file': file, 'vocal': vocal, 'emotion': emotion})

    # split to train, test sets using scikit-learn

    train, test = train_test_split(df, test_size=0.2, random_state=42)

    # save to csv
    train.to_csv('asvp_train.csv', index=False)
    test.to_csv('asvp_test.csv', index=False)

if __name__ == '__main__':
    main()



# model_svm.py

from sklearn import svm
from nkululeko.models.model import Model


class SVM_model(Model):
    """An SVM model"""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.name = "svm"
        c = float(self.util.config_val("MODEL", "C_val", "0.001"))
        if eval(self.util.config_val("MODEL", "class_weight", "False")):
            class_weight = "balanced"
        else:
            class_weight = None
        kernel = self.util.config_val("MODEL", "kernel", "rbf")
        self.clf = svm.SVC(
            kernel=kernel,
            C=c,
            gamma="scale",
            probability=True,
            class_weight=class_weight,
        )  # set up the classifier

    def set_c(self, c):
        """Set the C parameter."""
        self.clf.C = c

    def get_type(self):
        return "svm"



"""
This script is to import the 
The Ryerson Audio-Visual Database of Emotional Speech and Song (RAVDESS)
database to nkululeko.

I used the version downloadable from [Zenodo](https://zenodo.org/record/1188976)

Download and unzip the file Audio_Speech_Actors_01-24.zip e.g., `ravdess_speech`

adapted from https://www.kaggle.com/code/shivamburnwal/speech-emotion-recognition

Usage: `python3 process_database.py -d /data/ravdess_speech
OR
`python3 process_database.py /data/ravdess_speech`
"""

 
import argparse
import os
import sys
from pathlib import Path, PurePath

import pandas as pd

# ravdess source directory as argument
# source_dir = './'
parser = argparse.ArgumentParser()
parser.add_argument('-d', '--dir', type=str, default='./', help='path to RAVDESS speech directory')
args = parser.parse_args()
source_dir = str(Path(args.dir))
database_name = 'ravdess'

# check if directory (e.g., Actor_01) exists
if not os.path.exists(os.path.join(source_dir, "Actor_01")):
    print(f'Error: {source_dir}/Actor_01 does not exist')
    raise FileNotFoundError

ravdess_directory_list = os.listdir(source_dir)
# print(ravdess_directory_list)

file_emotion = []
file_speaker = []
file_gender = []
file_path = []
for dir in ravdess_directory_list:
    # as their are 20 different actors in our previous directory we need to extract files for each actor.
    if not dir.startswith('Actor'):
        continue
    if os.path.isdir(os.path.join(source_dir, dir)):
        actor = os.listdir(os.path.join(source_dir, dir))
        for file in actor:
            part = file.split('.')[0]
            part = part.split('-')
            # third part in each file represents the emotion associated to that file.
            file_emotion.append(int(part[2]))
            file_gender.append(int(part[6]))
            file_speaker.append(str(part[6]))
            file_path.append(source_dir + '/' + dir + '/' + file)
        
# put all in one dataframe
result_df = pd.DataFrame(list(zip(file_path, file_emotion,file_speaker,file_gender)), columns=['file','emotion', 'speaker', 'gender'])

# change speaker labels to string
# result_df['speaker'] = result_df['speaker'].astype('str')

# add `spk` to speaker id
result_df['speaker'] = 'spk' + result_df['speaker'].astype('str')

# changing integers to actual emotions.
result_df.emotion.replace({1:'neutral', 2:'calm', 3:'happy', 4:'sad', 5:'angry', 6:'fear', 7:'disgust', 8:'surprise'}, inplace=True)

# changing numbers to gender designations
def get_gender(x):
    if x%2==0:
        return 'female'
    else:
        return 'male'
result_df.gender = result_df.gender.map(lambda x: get_gender(x))

train_df, test_df = result_df.groupby('speaker').apply(lambda x: x.sample(frac=0.8)).reset_index(drop=True), result_df.groupby('speaker').apply(lambda x: x.drop(x.sample(frac=0.8).index)).reset_index(drop=True)

# save to csv
train_df.to_csv(f'{database_name}_speaker_train.csv', index=False)
test_df.to_csv(f'{database_name}_speaker_test.csv', index=False)

# print length of train and test set and total
print(f"Total length: {len(result_df)}, Train set: {len(train_df)}, Test set: {len(test_df)}")




#!/usr/bin/env python3

import argparse
import glob
import os

import pandas as pd


def main(data_dir):
    files = glob.glob(os.path.join(data_dir + 'wav/*/*/', '*.wav'))
    files.sort()

    data_train = []
    data_dev = []
    data_test = []
    
    for file in files:
        # processing file
        # print("Processing... ", file)
        lab_str = os.path.basename(os.path.dirname(file))
        # use text 1-30 as training, the rest as test        
        if int(os.path.basename(file)[8:10]) in range(1, 31):
            data_train.append({
                "file": file,
                "emotion": lab_str,
                "speaker": int(os.path.basename(file)[1:3]),
            })
    
        # use text 31-40 as dev, the rest as test
        elif int(os.path.basename(file)[8:10]) in range(31, 41):
            data_dev.append({
                "file": file,
                "emotion": lab_str,
                "speaker": int(os.path.basename(file)[1:3]),
            })
        
        else:
            data_test.append({
                "file": file,
                "emotion": lab_str,
                "speaker": int(os.path.basename(file)[1:3]),
            })
    # check length of data
    print(f"Lenghth train, dev, test: {len(data_train)}, {len(data_dev)}, {len(data_test)}")

    # save as csv
    pd.DataFrame(data_train).to_csv(f"jtes_ti_train.csv", index=False)
    pd.DataFrame(data_dev).to_csv(f"jtes_ti_dev.csv", index=False)
    pd.DataFrame(data_test).to_csv(f"jtes_ti_test.csv", index=False)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='./jtes_v1.1/')
    args = parser.parse_args()
    main(args.data_dir)



"""
This script is to import the 
The Ryerson Audio-Visual Database of Emotional Speech and Song (RAVDESS)
database to nkululeko.

I used the version downloadable from [Zenodo](https://zenodo.org/record/1188976)

Download and unzip the file Audio_Speech_Actors_01-24.zip e.g., `ravdess_speech`

adapted from https://www.kaggle.com/code/shivamburnwal/speech-emotion-recognition

<<<<<<< HEAD
Usage: `python3 process_database.py /data/ravdess_speech/`

=======
Usage: `python3 process_database.py -d /data/ravdess_speech
OR
`python3 process_database.py /data/ravdess_speech`
>>>>>>> e639b1104ec3f50b3944440119f2416aea6e31ac
"""

 
import os
import pandas as pd
import sys
<<<<<<< HEAD

# ravdess source directory as argument
# source_dir = './'
source_dir = sys.argv[1]
=======
import argparse
from pathlib import Path, PurePath

# ravdess source directory as argument
# source_dir = './'
parser = argparse.ArgumentParser()
parser.add_argument('-d', '--dir', type=str, default='./', help='path to RAVDESS speech directory')
args = parser.parse_args()
source_dir = str(Path(args.dir))
>>>>>>> e639b1104ec3f50b3944440119f2416aea6e31ac
database_name = 'ravdess'

# check if directory (e.g., Actor_01) exists
if not os.path.exists(os.path.join(source_dir, "Actor_01")):
    print(f'Error: {source_dir}/Actor_01 does not exist')
    raise FileNotFoundError

ravdess_directory_list = os.listdir(source_dir)
print(ravdess_directory_list)

file_emotion = []
file_speaker = []
file_gender = []
file_path = []
for dir in ravdess_directory_list:
    # as their are 20 different actors in our previous directory we need to extract files for each actor.
    if not dir.startswith('Actor'):
        continue
    if os.path.isdir(os.path.join(source_dir, dir)):
        actor = os.listdir(os.path.join(source_dir, dir))
        for file in actor:
            part = file.split('.')[0]
            part = part.split('-')
            # third part in each file represents the emotion associated to that file.
            file_emotion.append(int(part[2]))
            file_gender.append(int(part[6]))
            file_speaker.append(str(part[6]))
            file_path.append(source_dir + '/' + dir + '/' + file)
        
# put all in one dataframe
result_df = pd.DataFrame(list(zip(file_path, file_emotion,file_speaker,file_gender)), columns=['file','emotion', 'speaker', 'gender'])


# changing integers to actual emotions.
result_df.emotion.replace({1:'neutral', 2:'calm', 3:'happy', 4:'sad', 5:'angry', 6:'fear', 7:'disgust', 8:'surprise'}, inplace=True)
# changing numbers to gender designations
def get_gender(x):
    if x%2==0:
        return 'female'
    else:
        return 'male'
result_df.gender = result_df.gender.map(lambda x: get_gender(x))

# For a gender balanced split close to 60%-20%-20%: train - 16, dev - 4 , test - 4
speaker_splits = {
    'train': ['01', '02', '03', '04', '05', '06', '07', '08', '09', '10', '11', '12', '13', '14', '15', '18'],
    'dev': ['16', '17', '19', '20'],
    'test': ['21', '22', '23', '24']
}

# set the index to file paths
result_df = result_df.set_index('file')

# split the data
for split in ['train','test','dev']:
    split_df = result_df[result_df.speaker.isin(speaker_splits[split])]
    split_df.to_csv(f'{database_name}_{split}.csv')
    print(f'{split} split #samples: {split_df.shape[0]}')


# make a convinience file with all data
result_df.to_csv(database_name+'.csv')


"""
Should result into something like
file,emotion,speaker,gender
./Actor_09/03-01-08-01-02-02-09.wav,surprise,ravdess_09,male
./Actor_09/03-01-08-02-01-01-09.wav,surprise,ravdess_09,male
...
"""



"""
Code based on @jwagner
"""

import dataclasses
import typing

import torch
import transformers
from transformers.models.wav2vec2.modeling_wav2vec2 import (
    Wav2Vec2PreTrainedModel,
    Wav2Vec2Model,
)


class ConcordanceCorCoeff(torch.nn.Module):

    def __init__(self):

        super().__init__()

        self.mean = torch.mean
        self.var = torch.var
        self.sum = torch.sum
        self.sqrt = torch.sqrt
        self.std = torch.std

    def forward(self, prediction, ground_truth):

        mean_gt = self.mean(ground_truth, 0)
        mean_pred = self.mean(prediction, 0)
        var_gt = self.var(ground_truth, 0)
        var_pred = self.var(prediction, 0)
        v_pred = prediction - mean_pred
        v_gt = ground_truth - mean_gt
        cor = self.sum(v_pred * v_gt) / (
            self.sqrt(self.sum(v_pred**2)) * self.sqrt(self.sum(v_gt**2))
        )
        sd_gt = self.std(ground_truth)
        sd_pred = self.std(prediction)
        numerator = 2 * cor * sd_gt * sd_pred
        denominator = var_gt + var_pred + (mean_gt - mean_pred) ** 2
        ccc = numerator / denominator

        return 1 - ccc


@dataclasses.dataclass
class ModelOutput(transformers.file_utils.ModelOutput):

    logits_cat: torch.FloatTensor = None
    hidden_states: typing.Tuple[torch.FloatTensor] = None
    cnn_features: torch.FloatTensor = None


class ModelHead(torch.nn.Module):

    def __init__(self, config, num_labels):

        super().__init__()

        self.dense = torch.nn.Linear(config.hidden_size, config.hidden_size)
        self.dropout = torch.nn.Dropout(config.final_dropout)
        self.out_proj = torch.nn.Linear(config.hidden_size, num_labels)

    def forward(self, features, **kwargs):

        x = features
        x = self.dropout(x)
        x = self.dense(x)
        x = torch.tanh(x)
        x = self.dropout(x)
        x = self.out_proj(x)

        return x


class Model(Wav2Vec2PreTrainedModel):

    def __init__(self, config):

        super().__init__(config)

        self.wav2vec2 = Wav2Vec2Model(config)
        self.cat = ModelHead(config, 2)
        self.init_weights()

    def freeze_feature_extractor(self):
        self.wav2vec2.feature_extractor._freeze_parameters()

    def pooling(
        self,
        hidden_states,
        attention_mask,
    ):

        if attention_mask is None:  # For evaluation with batch_size==1
            outputs = torch.mean(hidden_states, dim=1)
        else:
            attention_mask = self._get_feature_vector_attention_mask(
                hidden_states.shape[1],
                attention_mask,
            )
            hidden_states = hidden_states * torch.reshape(
                attention_mask,
                (-1, attention_mask.shape[-1], 1),
            )
            outputs = torch.sum(hidden_states, dim=1)
            attention_sum = torch.sum(attention_mask, dim=1)
            outputs = outputs / torch.reshape(attention_sum, (-1, 1))

        return outputs

    def forward(
        self,
        input_values,
        attention_mask=None,
        labels=None,
        return_hidden=False,
    ):

        outputs = self.wav2vec2(
            input_values,
            attention_mask=attention_mask,
        )

        cnn_features = outputs.extract_features
        hidden_states_framewise = outputs.last_hidden_state
        hidden_states = self.pooling(
            hidden_states_framewise,
            attention_mask,
        )
        logits_cat = self.cat(hidden_states)

        if not self.training:
            logits_cat = torch.softmax(logits_cat, dim=1)

        if return_hidden:

            # make time last axis
            cnn_features = torch.transpose(cnn_features, 1, 2)

            return ModelOutput(
                logits_cat=logits_cat,
                hidden_states=hidden_states,
                cnn_features=cnn_features,
            )

        else:

            return ModelOutput(
                logits_cat=logits_cat,
            )

    def predict(self, signal):
        result = self(torch.from_numpy(signal))
        result = result[0].detach().numpy()[0]
        return result


class ModelWithPreProcessing(Model):

    def __init__(self, config):
        super().__init__(config)

    def forward(
        self,
        input_values,
    ):
        # Wav2Vec2FeatureExtractor.zero_mean_unit_var_norm():
        # normed_slice = (vector - vector[:length].mean()) / np.sqrt(vector[:length].var() + 1e-7)

        mean = input_values.mean()

        # var = input_values.var()
        # raises: onnxruntime.capi.onnxruntime_pybind11_state.NotImplemented: [ONNXRuntimeError] : 9 : NOT_IMPLEMENTED : Could not find an implementation for the node ReduceProd_3:ReduceProd(11)

        var = torch.square(input_values - mean).mean()
        input_values = (input_values - mean) / torch.sqrt(var + 1e-7)

        output = super().forward(
            input_values,
            return_hidden=True,
        )

        return (
            output.hidden_states,
            output.logits_cat,
            output.cnn_features,
        )



# model_gmm.py

from sklearn import mixture
from nkululeko.models.model import Model


class GMM_model(Model):
    """An GMM model"""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.name = "gmm"
        n_components = int(self.util.config_val("MODEL", "GMM_components", "4"))
        covariance_type = self.util.config_val("MODEL", "GMM_covariance_type", "full")
        self.clf = mixture.GaussianMixture(
            n_components=n_components, covariance_type=covariance_type
        )
        # set up the classifier



""""
A predictor for MOS - mean opinion score.
"""
from nkululeko.utils.util import Util
import ast
import nkululeko.glob_conf as glob_conf
from nkululeko.feature_extractor import FeatureExtractor
import numpy as np


class MOSPredictor:
    """
    MOSPredictor
    predicting MOS

    """

    def __init__(self, df):
        self.df = df
        self.util = Util("mosPredictor")

    def predict(self, split_selection):
        self.util.debug(f"estimating MOS for {split_selection} samples")
        return_df = self.df.copy()
        feats_name = "_".join(ast.literal_eval(glob_conf.config["DATA"]["databases"]))
        self.feature_extractor = FeatureExtractor(
            self.df, ["mos"], feats_name, split_selection
        )
        result_df = self.feature_extractor.extract()
        # replace missing values by 0
        result_df = result_df.fillna(0)
        result_df = result_df.replace(np.nan, 0)
        result_df.replace([np.inf, -np.inf], 0, inplace=True)
        pred_snr = result_df.mos * 100
        return_df["mos_pred"] = pred_snr.astype("int") / 100
        return return_df



#!/usr/bin/env python3
# process_database.py --> EMORYNLP

import argparse
from pathlib import Path

import pandas as pd

NAME_FMT = "sea{0.Season}_ep{0.Episode}_sc{0.Scene_ID}_utt{0.Utterance_ID}"

emotion_map = {
    "Joyful": "happiness",
    "Mad": "anger",
    "Neutral": "neutral",
    "Peaceful": "peace",
    "Powerful": "power",
    "Sad": "sadness",
    "Scared": "fear",
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="EMORYNLP")
    parser.add_argument("--output_dir", type=str, default=".")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    splits = ["train", "dev", "test"]

    for split in splits:
        input_csv_file = data_dir / f"emorynlp_{split}_final.csv"
        df = pd.read_csv(input_csv_file)
        df.index = df.apply(lambda x: NAME_FMT.format(x), axis=1)
        df["split"] = split
        if split == "train":
            df = df.drop(
                ["sea4_ep4_sc3_utt10", "sea4_ep4_sc3_utt8", "sea4_ep4_sc3_utt9"])

        df["Emotion"] = df["Emotion"].map(emotion_map)
        df = df.rename(columns={"Emotion": "emotion",
                       "Sentiment": "sentiment", "Speaker": "speaker"})
        df.index = df.index + ".wav"
        df.index.name = "file"
        df.to_csv(output_dir / f"emorynlp_{split}.csv")
        df.index = df.index.str.replace(".wav", ".mp4")
        df.to_csv(output_dir / f"emorynlp_{split}_mp4.csv")

        # print number of each split
        print(f"{split}: {len(df)}")


if __name__ == "__main__":
    main()



"""test_predictor.py.

    Predict targets from a model and save as csv file.

"""

import ast

import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder

from nkululeko.data.dataset import Dataset
from nkululeko.feature_extractor import FeatureExtractor
import nkululeko.glob_conf as glob_conf
from nkululeko.scaler import Scaler
from nkululeko.utils.util import Util


class TestPredictor:
    def __init__(self, model, orig_df, labenc, name):
        """Constructor setting up name and configuration."""
        self.model = model
        self.orig_df = orig_df
        self.label_encoder = labenc
        self.target = glob_conf.config["DATA"]["target"]
        self.util = Util("test_predictor")
        self.name = name

    def predict_and_store(self):
        label_data = self.util.config_val("DATA", "label_data", False)
        result = 0
        if label_data:
            data = Dataset(label_data)
            data.load()
            data.prepare_labels()
            data_df = self.util.make_segmented_index(data.df)
            data_df.is_labeled = data.is_labeled

            featextractor = FeatureExtractor(data_df, label_data, "")
            feats_df = featextractor.extract()
            scale = self.util.config_val("FEATS", "scale", False)
            labelenc = LabelEncoder()
            data_df[self.target] = labelenc.fit_transform(data_df[self.target])
            #            data_df[self.target] = self.label_encoder.fit_transform(data_df[self.target])
            if scale:
                self.scaler = Scaler(data_df, None, feats_df, None, scale)
                feats_df, _ = self.scaler.scale()
            self.model.set_testdata(data_df, feats_df)
            predictions = self.model.get_predictions()
            df = pd.DataFrame(index=data_df.index)
            df["speaker"] = data_df["speaker"]
            df["gender"] = data_df["gender"]
            df[self.target] = labelenc.inverse_transform(predictions.tolist())
            df.to_csv(self.name)
        else:
            test_dbs = ast.literal_eval(glob_conf.config["DATA"]["tests"])
            test_dbs_string = "_".join(test_dbs)
            predictions = self.model.get_predictions()
            report = self.model.predict()
            result = report.result.get_result()
            report.set_filename_add(f"test-{test_dbs_string}")
            self.util.print_best_results([report])
            report.plot_confmatrix(self.util.get_plot_name(), 0)
            report.print_results(0)
            # print(predictions)
            # df = pd.DataFrame(index=self.orig_df.index)
            # df["speaker"] = self.orig_df["speaker"]
            # df["gender"] = self.orig_df["gender"]
            # df[self.target] = self.orig_df[self.target]
            df = self.orig_df.copy()
            df["predictions"] = self.label_encoder.inverse_transform(predictions)
            target = self.util.config_val("DATA", "target", "emotion")
            if "class_label" in df.columns:
                df = df.drop(columns=[target])
                df = df.rename(columns={"class_label": target})
            df.to_csv(self.name)
        self.util.debug(f"results stored in {self.name}")
        return result



# util.py
import ast
import configparser
import os.path
import pickle
import sys

import numpy as np
import pandas as pd

import audeer
import audformat


class Util:
    # a list of words that need not to be warned upon if default values are used
    stopvals = [
        "all",
        False,
        "False",
        "classification",
        "png",
        "audio_path",
        "kde",
        "pkl",
        "eGeMAPSv02",
        "functionals",
    ]

    def __init__(self, caller=None, has_config=True):
        if caller is not None:
            self.caller = caller
        else:
            self.caller = ""
        if has_config:
            try:
                import nkululeko.glob_conf as glob_conf
                self.config = glob_conf.config
                self.got_data_roots = self.config_val(
                    "DATA", "root_folders", False)
                if self.got_data_roots:
                    # if there is a global data rootfolder file, read from there
                    if not os.path.isfile(self.got_data_roots):
                        self.error(f"no such file: {self.got_data_roots}")
                    self.data_roots = configparser.ConfigParser()
                    self.data_roots.read(self.got_data_roots)
            except (ModuleNotFoundError, AttributeError):
                self.config = None
                self.got_data_roots = False

    def get_path(self, entry):
        """
        This method allows the user to get the directory path for the given argument.
        """
        if self.config is None:
            # If no configuration file is provided, use default paths
            if entry == "fig_dir":
                dir_name = "./images/"
            elif entry == "res_dir":
                dir_name = "./results/"
            elif entry == "model_dir":
                dir_name = "./models/"
            else:
                dir_name = "./store/"
        else:
            root = os.path.join(self.config["EXP"]["root"], "")
            name = self.config["EXP"]["name"]
            try:
                entryn = self.config["EXP"][entry]
            except KeyError:
                # some default values
                if entry == "fig_dir":
                    entryn = "./images/"
                elif entry == "res_dir":
                    entryn = "./results/"
                elif entry == "model_dir":
                    entryn = "./models/"
                else:
                    entryn = "./store/"

            # Expand image, model and result directories with run index
            if entry == "fig_dir" or entry == "res_dir" or entry == "model_dir":
                run = self.config_val("EXP", "run", 0)
                entryn = entryn + f"run_{run}/"

            dir_name = f"{root}{name}/{entryn}"

        audeer.mkdir(dir_name)
        return dir_name

    def config_val_data(self, dataset, key, default):
        """
        Retrieve a configuration value for datasets.
        If the value is present in the experiment configuration it will be used, else
        we look in a global file specified by the root_folders value.
        """
        import nkululeko.glob_conf as glob_conf

        configuration = glob_conf.config
        try:
            if len(key) > 0:
                return configuration["DATA"][dataset + "." + key].strip("'\"")
            else:
                return configuration["DATA"][dataset].strip("'\"")
        except KeyError:
            if self.got_data_roots:
                try:
                    if len(key) > 0:
                        return self.data_roots["DATA"][dataset + "." + key].strip("'\"")
                    else:
                        return self.data_roots["DATA"][dataset].strip("'\"")
                except KeyError:
                    if not default in self.stopvals:
                        self.debug(
                            f"value for {key} not found, using default:" f" {default}"
                        )
                    return default
            if not default in self.stopvals:
                self.debug(
                    f"value for {key} not found, using default: {default}")
            return default

    def set_config(self, config):
        self.config = config

    def get_save_name(self):
        """Return a relative path to a name to save the experiment"""
        store = self.get_path("store")
        return f"{store}/{self.get_exp_name()}.pkl"

    def is_categorical(self, pd_series):
        """Check if a dataframe column is categorical"""
        return pd_series.dtype.name == "object" or isinstance(
            pd_series.dtype, pd.CategoricalDtype
        )

    def get_exp_dir(self):
        """
        Get the experiment directory
        """
        root = os.path.join(self.config["EXP"]["root"], "")
        name = self.config["EXP"]["name"]
        dir_name = f"{root}{name}"
        audeer.mkdir(dir_name)
        return dir_name

    def get_res_dir(self):
        home_dir = self.get_exp_dir()
        dir_name = f"{home_dir}/results/"
        audeer.mkdir(dir_name)
        return dir_name

    def make_segmented_index(self, df):
        if len(df) == 0:
            return df
        if not isinstance(df.index, pd.MultiIndex):
            df.index = audformat.utils.to_segmented_index(
                df.index, allow_nat=False)
        return df

    def _get_value_descript(self, section, name):
        if self.config_val(section, name, False):
            val = self.config_val(section, name, False)
            val = str(val).strip(".")
            return f"_{name}-{str(val)}"
        return ""

    def get_data_name(self):
        """
        Get a string as name from all databases that are useed
        """
        return "_".join(ast.literal_eval(self.config["DATA"]["databases"]))

    def get_feattype_name(self):
        """
        Get a string as name from all feature sets that are useed
        """
        return "_".join(ast.literal_eval(self.config["FEATS"]["type"]))

    def get_exp_name(self, only_train=False, only_data=False):
        trains_val = self.config_val("DATA", "trains", False)
        if only_train and trains_val:
            # try to get only the train tables
            ds = "_".join(ast.literal_eval(self.config["DATA"]["trains"]))
        else:
            ds = "_".join(ast.literal_eval(self.config["DATA"]["databases"]))
        return_string = f"{ds}"
        if not only_data:
            mt = self.get_model_description()
            return_string = return_string + "_" + mt
        return return_string.replace("__", "_")

    def get_model_description(self):
        mt = ""
        mt = f'{self.config["MODEL"]["type"]}'
        ft = "_".join(ast.literal_eval(self.config["FEATS"]["type"]))
        ft += "_"
        layer_string = ""
        layer_s = self.config_val("MODEL", "layers", False)
        if layer_s:
            layers = ast.literal_eval(layer_s)
            sorted_layers = sorted(layers.items(), key=lambda x: x[1])
            for l in sorted_layers:
                layer_string += f"{str(l[1])}-"
        return_string = f"{mt}_{ft}{layer_string[:-1]}"
        options = [
            ["MODEL", "C_val"],
            ["MODEL", "kernel"],
            ["MODEL", "drop"],
            ["MODEL", "class_weight"],
            ["MODEL", "loss"],
            ["MODEL", "logo"],
            ["MODEL", "learning_rate"],
            ["MODEL", "k_fold_cross"],
            ["FEATS", "balancing"],
            ["FEATS", "scale"],
            ["FEATS", "set"],
            ["FEATS", "wav2vec2.layer"],
        ]
        for option in options:
            return_string += self._get_value_descript(option[0], option[1]).replace(
                ".", "-"
            )
        return return_string

    def get_plot_name(self):
        try:
            plot_name = self.config["PLOT"]["name"]
        except KeyError:
            plot_name = self.get_exp_name()
        return plot_name

    def exp_is_classification(self):
        type = self.config_val("EXP", "type", "classification")
        if type == "classification":
            return True
        return False

    def error(self, message):
        print(f"ERROR {self.caller}: {message}")
        sys.exit()

    def warn(self, message):
        print(f"WARNING {self.caller}: {message}")

    def debug(self, message):
        print(f"DEBUG {self.caller}: {message}")

    def set_config_val(self, section, key, value):
        try:
            # does the section already exists?
            self.config[section][key] = str(value)
        except KeyError:
            self.config.add_section(section)
            self.config[section][key] = str(value)

    def check_df(self, i, df):
        """Check a dataframe"""
        print(f"check {i}: {df.shape}")
        print(df.head(1))

    def config_val(self, section, key, default):
        if self.config is None:
            return default
        try:
            return self.config[section][key]
        except KeyError:
            if default not in self.stopvals:
                self.debug(
                    f"value for {key} not found, using default: {default}")
            return default

    def config_val_list(self, section, key, default):
        try:
            return ast.literal_eval(self.config[section][key])
        except KeyError:
            if not default in self.stopvals:
                self.debug(
                    f"value for {key} not found, using default: {default}")
            return default

    def continuous_to_categorical(self, series):
        """
        discretize a categorical variable.
        uses the labels and bins from the ini if present

        :param series: a pandas series
        :return a pandas series with discretized values as categories
        """
        try:
            bins = ast.literal_eval(self.config["DATA"]["bins"])
            labels = ast.literal_eval(self.config["DATA"]["labels"])
        except KeyError:
            # if no binning is given, simply take three bins
            b1 = np.quantile(series, 0.33)
            b2 = np.quantile(series, 0.66)
            bins = [-1000000, b1, b2, 1000000]
            labels = ["0_low", "1_middle", "2_high"]
        result = np.digitize(series, bins) - 1
        result = pd.Series(result)
        for i, l in enumerate(labels):
            result = result.replace(i, str(l))
        result = result.astype("category")
        return result

    def print_best_results(self, best_reports):
        res_dir = self.get_res_dir()
        # go one level up above the "run" level
        all = ""
        vals = np.empty(0)
        for report in best_reports:
            all += str(report.result.test) + ", "
            vals = np.append(vals, report.result.test)
        file_name = f"{res_dir}{self.get_exp_name()}_runs.txt"
        with open(file_name, "w") as text_file:
            text_file.write(all)
            text_file.write(
                f"\nmean: {vals.mean():.3f}, max: {vals.max():.3f}, max_index:"
                f" {vals.argmax()}"
            )

    def exist_pickle(self, name):
        store = self.get_path("store")
        name = "/".join([store, name]) + ".pkl"
        if os.path.isfile(name):
            return True
        return False

    def to_pickle(self, anyobject, name):
        store = self.get_path("store")
        name = "/".join([store, name]) + ".pkl"
        self.debug(f"saving {name}")
        with open(name, "wb") as handle:
            pickle.dump(anyobject, handle)

    def from_pickle(self, name):
        store = self.get_path("store")
        name = "/".join([store, name]) + ".pkl"
        self.debug(f"loading {name}")
        with open(name, "rb") as handle:
            any_opject = pickle.load(handle)
        return any_opject

    def write_store(self, df, storage, format):
        if format == "pkl":
            df.to_pickle(storage)
        elif format == "csv":
            df.to_csv(storage)
        else:
            self.error(f"unkown store format: {format}")

    def get_store(self, name, format):
        if format == "pkl":
            return pd.read_pickle(name)
        elif format == "csv":
            return audformat.utils.read_csv(name)
        else:
            self.error(f"unknown store format: {format}")

    def save_to_store(self, df, name):
        store = self.get_path("store")
        store_format = self.config_val("FEATS", "store_format", "pkl")
        storage = f"{store}{name}.{store_format}"
        self.write_store(df, storage, store_format)

    def copy_flags(self, df_source, df_target):
        if hasattr(df_source, "is_labeled"):
            df_target.is_labeled = df_source.is_labeled
        if hasattr(df_source, "is_test"):
            df_target.is_test = df_source.is_test
        if hasattr(df_source, "is_train"):
            df_target.is_train = df_source.is_train
        if hasattr(df_source, "is_val"):
            df_target.is_val = df_source.is_val
        if hasattr(df_source, "got_gender"):
            df_target.got_gender = df_source.got_gender
        if hasattr(df_source, "got_age"):
            df_target.got_age = df_source.got_age
        if hasattr(df_source, "got_speaker"):
            df_target.got_speaker = df_source.got_speaker

    def high_is_good(self):
        """check how to interpret results"""
        if self.exp_is_classification():
            return True
        else:
            measure = self.config_val("MODEL", "measure", "mse")
            if measure == "mse" or measure == "mae":
                return False
            elif measure == "ccc":
                return True
            else:
                self.error(f"unknown measure: {measure}")

    def to_3_digits(self, x):
        x = float(x)
        return (int(x * 1000)) / 1000.0



# svrmodel.py

from sklearn import svm
from nkululeko.models.model import Model


class SVR_model(Model):
    """An SVR model"""

    is_classifier = False

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.name = "svr"
        c = float(self.util.config_val("MODEL", "C_val", "0.001"))
        #  kernel{‘linear’, ‘poly’, ‘rbf’, ‘sigmoid’, ‘precomputed’} or callable, default=’rbf’
        kernel = self.util.config_val("MODEL", "kernel", "rbf")
        self.clf = svm.SVR(kernel=kernel, C=c)  # set up the classifier

    def set_C(self, c):
        """Set the C parameter"""
        self.clf.C = c



# result.py


class Result:
    def __init__(self, test, train, loss, loss_eval, measure):
        self.test = test
        self.train = train
        self.loss = loss
        self.loss_eval = loss_eval
        self.measure = measure

    def get_result(self):
        return self.test

    def set_upper_lower(self, upper, lower):
        """Set the upper and lower bound of confidence interval."""
        self.upper = upper
        self.lower = lower

    def get_test_result(self):
        return f"test: {self.test:.3f} {self.measure}"

    def to_string(self):
        return (
            f"test: {self.test} {self.measure}, train:"
            f" {self.train} {self.measure}, loss: {self.loss}, eval-loss: {self.loss_eval}"
        )



# test.py
# Just use a database as test

import argparse
import configparser
import os

from nkululeko.constants import VERSION
from nkululeko.experiment import Experiment
from nkululeko.utils.util import Util


def do_it(config_file, outfile):

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)

    # create a new experiment
    expr = Experiment(config)
    module = "test"
    expr.set_module(module)
    util = Util(module, has_config=True)
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    # load the experiment
    expr.load(f"{util.get_save_name()}")
    expr.fill_tests()
    expr.extract_test_feats()
    result = expr.predict_test_and_save(outfile)

    print("DONE")

    return result, 0


def main(src_dir):
    parser = argparse.ArgumentParser(description="Call the nkululeko TEST framework.")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    parser.add_argument(
        "--outfile",
        default="my_results.csv",
        help="File name to store the predictions",
    )
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"
    do_it(config_file, args.outfile)


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # use this if you want to state the config file path on command line



# feats_whisper.py
import os

import pandas as pd
import torch
from transformers import AutoFeatureExtractor
from transformers import WhisperModel

import audeer
import audiofile

from nkululeko.feat_extract.featureset import Featureset
import nkululeko.glob_conf as glob_conf


class Whisper(Featureset):
    """Class to extract whisper embeddings."""

    def __init__(self, name, data_df, feat_type):
        super().__init__(name, data_df, feat_type)
        cuda = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = self.util.config_val("MODEL", "device", cuda)
        self.model_initialized = False
        if feat_type == "whisper":
            self.feat_type = "whisper-base"
        else:
            self.feat_type = feat_type

    def init_model(self):
        # load model
        self.util.debug("loading whisper model...")
        model_name = f"openai/{self.feat_type}"
        self.model = WhisperModel.from_pretrained(model_name).to(self.device)
        print(f"intialized Whisper model on {self.device}")
        self.feature_extractor = AutoFeatureExtractor.from_pretrained(model_name)
        self.model_initialized = True

    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path("store")
        storage = f"{store}{self.name}.pkl"
        extract = self.util.config_val("FEATS", "needs_feature_extraction", False)
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            if not self.model_initialized:
                self.init_model()
            self.util.debug("extracting whisper embeddings, this might take a while...")
            emb_series = []
            for (file, start, end), _ in audeer.progress_bar(
                self.data_df.iterrows(),
                total=len(self.data_df),
                desc=f"Running whisper on {len(self.data_df)} audiofiles",
            ):
                if end == pd.NaT:
                    signal, sr = audiofile.read(file, offset=start)
                else:
                    signal, sr = audiofile.read(
                        file, duration=end - start, offset=start
                    )
                emb = self.get_embeddings(signal, sr, file)
                emb_series.append(emb)
            # print(f"emb_series shape: {emb_series.shape}")
            self.df = pd.DataFrame(emb_series, index=self.data_df.index)
            # print(f"df shape: {self.df.shape}")
            self.df.to_pickle(storage)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug("reusing extracted wav2vec2 embeddings")
            self.df = pd.read_pickle(storage)
            if self.df.isnull().values.any():
                nanrows = self.df.columns[self.df.isna().any()].tolist()
                # print(nanrows)
                self.util.error(
                    f"got nan: {self.df.shape} {self.df.isnull().sum().sum()}"
                )

    def get_embeddings(self, signal, sampling_rate, file):
        r"""Extract embeddings from raw audio signal."""
        try:
            with torch.no_grad():
                embed_size = self.model.config.hidden_size
                embed_columns = [f"whisper_{i}" for i in range(embed_size)]
                inputs = self.feature_extractor(signal, sampling_rate=16000)[
                    "input_features"
                ][0]
                inputs = torch.from_numpy(inputs).to(self.device).unsqueeze(0)
                decoder_input_ids = (
                    torch.tensor([[1, 1]]).to(self.device)
                    * self.model.config.decoder_start_token_id
                )
                full_outputs = self.model(
                    inputs,
                    decoder_input_ids=decoder_input_ids,
                    output_hidden_states=True,
                )
                outputs = full_outputs.encoder_last_hidden_state[0]
                average_embeds = outputs.squeeze().mean(axis=0).cpu().detach().numpy()
        except RuntimeError as re:
            print(str(re))
            self.util.error(f"couldn't extract file: {file}")
        # print(f"y flattened shape: {y.ravel().shape}")
        return average_embeds

    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_embeddings(signal, sr, "no file")
        return feats



"""Runmanager module.

This module contains the Runmanager class which is responsible for managing the 
runs of the experiment.
"""

import nkululeko.glob_conf as glob_conf
from nkululeko.modelrunner import Modelrunner
from nkululeko.reporting.reporter import Reporter
from nkululeko.utils.util import Util


class Runmanager:
    """Class to manage the runs of the experiment (e.g. when results differ caused by random initialization)"""

    model = None  # The underlying model
    df_train, df_test, feats_train, feats_test = (
        None,
        None,
        None,
        None,
    )  # The dataframes
    reports = []

    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor setting up the dataframes
        Args:
            df_train: train dataframe
            df_test: test dataframe
            feats_train: train features
            feats_train: test features

        Returns:

        """
        self.df_train, self.df_test, self.feats_train, self.feats_test = (
            df_train,
            df_test,
            feats_train,
            feats_test,
        )
        self.util = Util("runmanager")
        self.target = glob_conf.config["DATA"]["target"]
        # intialize a new model
        # model_type = glob_conf.config['MODEL']['type']
        # self._select_model(model_type)

    def do_runs(self):
        """Start the runs"""
        self.best_results = []  # keep the best result per run
        self.last_epochs = []  # keep the epoch of best result per run
        # for all runs
        for run in range(int(self.util.config_val("EXP", "runs", 1))):
            self.util.debug(f"run {run}")
            # set the run index as global variable for reporting
            self.util.set_config_val("EXP", "run", run)
            self.modelrunner = Modelrunner(
                self.df_train,
                self.df_test,
                self.feats_train,
                self.feats_test,
                run,
            )
            self.reports, last_epoch = self.modelrunner.do_epochs()
            # wrap up the run
            plot_anim_progression = self.util.config_val(
                "PLOT", "anim_progression", 0)
            if plot_anim_progression:
                plot_name_suggest = self.util.get_exp_name()
                plot_name = (
                    self.util.config_val("PLOT", "name", plot_name_suggest)
                    + "_conf_anim.gif"
                )
                self.util.debug(f"plotting animated confusion to {plot_name}")
                self.reports[-1].make_conf_animation(plot_name)
            plot_epoch_progression = self.util.config_val(
                "PLOT", "epoch_progression", 0
            )
            try:
                epoch_num = int(glob_conf.config["EXP"]["epochs"])
            except KeyError:
                # possibly this value has not been set
                epoch_num = 1
            if epoch_num > 1 and plot_epoch_progression:
                plot_name_suggest = self.util.get_exp_name()
                plot_name = (
                    self.util.config_val("PLOT", "name", plot_name_suggest)
                    + "_epoch_progression"
                )
                self.util.debug(f"plotting progression to {plot_name}")
                self.reports[-1].plot_epoch_progression(
                    self.reports, plot_name)
            # remember the best run
            best_report = self.get_best_result(self.reports)
            plot_best_model = self.util.config_val("PLOT", "best_model", False)

            if epoch_num > 1 and plot_best_model:
                plot_name_suggest = self.util.get_exp_name()
                plot_name = (
                    self.util.config_val("PLOT", "name", plot_name_suggest)
                    + f"_BEST_{best_report.run}_{best_report.epoch:03d}_BEST_cnf"
                )
                self.util.debug(
                    f"best result with run {best_report.run} and epoch"
                    f" {best_report.epoch}:"
                    f" {best_report.result.get_test_result()}"
                )
                self.print_model(best_report, plot_name)
            # finally, print out the numbers for this run
            self.reports[-1].print_results(
                int(self.util.config_val("EXP", "epochs", 1))
            )
            self.best_results.append(best_report)
            self.last_epochs.append(last_epoch)

    def print_best_result_runs(self):
        """Print the best result for all runs"""
        best_report = self.get_best_result(self.best_results)
        self.util.debug(
            f"best result all runs with run {best_report.run}             and"
            f" epoch {best_report.epoch}: {best_report.result.test:.3f}"
        )
        plot_name_suggest = self.util.get_exp_name()
        plot_name = (
            self.util.config_val("PLOT", "name", plot_name_suggest)
            + f"_BEST_{best_report.run}_{best_report.epoch:03d}_BEST_cnf"
        )
        self.print_model(best_report, plot_name)

    def print_given_result(self, run, epoch):
        """Print a result (confusion matrix) for a given epoch and run.

        Args:
            run: for which run
            epoch: for which epoch

        """
        report = Reporter([], [])
        report.set_id(run, epoch)
        self.util.debug(f"Re-testing result with run {run} and epoch {epoch}")
        plot_name_suggest = self.util.get_exp_name()
        plot_name = (
            self.util.config_val("PLOT", "name", plot_name_suggest)
            + f"_extra_{run}_{epoch:03d}_cnf"
        )
        self.print_model(report, plot_name)

    def print_model(self, report, plot_name):
        """Print a confusion matrix for a special report.

        Args:
            report: for which report (will be computed newly from model)
            plot_name: name of plot file
        """
        epoch = report.epoch
        # self.load_model(report)
        # report = self.model.predict()
        self.util.debug(f"plotting conf matrix to {plot_name}")
        report.plot_confmatrix(plot_name, epoch)
        report.print_results(epoch)

    def load_model(self, report):
        """Load a model from disk for a specific run and epoch and evaluate it.

        Args:
            report: for which report (will be re-evaluated)

        """
        run = report.run
        epoch = report.epoch
        self.util.set_config_val("EXP", "run", run)
        model_type = glob_conf.config["MODEL"]["type"]
        model = self.modelrunner._select_model(model_type)
        model.load(run, epoch)
        return model

    def get_best_model(self):
        best_report = self.get_best_result(self.best_results)
        return self.load_model(best_report)

    def get_best_result(self, reports):
        best_r = Reporter([], [], 0, 0)
        if self.util.high_is_good():
            best_r = self.search_best_result(reports, "ascending")
        else:
            best_r = self.search_best_result(reports, "descending")
        return best_r

    def search_best_result(self, reports, order):
        best_r = Reporter([], [], 0, 0)
        if order == "ascending":
            best_result = 0
            for r in reports:
                res = r.result.test
                if res > best_result:
                    best_result = res
                    best_r = r
        else:
            best_result = 10000
            for r in reports:
                res = r.result.test
                if res < best_result:
                    best_result = res
                    best_r = r
        return best_r



# demo.py
# Demonstration code to use the ML-experiment framework
# Test the loading of a previously trained model and demo mode
# needs the project config file to run before
"""This script is used to test the loading of a previously trained model.

And run it in demo mode.
It requires the project config file to be run before.

Usage:  
python -m nkululeko.demo [--config CONFIG] [--file FILE] [--list LIST] [--folder FOLDER] [--outfile OUTFILE]  

Options:   \n
--config CONFIG     The base configuration file (default: exp.ini) \n  
--file FILE         A file that should be processed (16kHz mono wav) \n  
--list LIST         A file with a list of files, one per line, that should be processed (16kHz mono wav) \n  
--folder FOLDER     A name of a folder where the files within the list are in   (default: ./) \n   
--outfile OUTFILE   A filename to store the results in CSV  (default: None)  
"""
import argparse
import configparser
import os

from nkululeko.constants import VERSION
from nkululeko.experiment import Experiment
import nkululeko.glob_conf as glob_conf
from nkululeko.utils.util import Util


def main(src_dir):
    parser = argparse.ArgumentParser(description="Call the nkululeko DEMO framework.")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    parser.add_argument(
        "--file", help="A file that should be processed (16kHz mono wav)"
    )
    parser.add_argument(
        "--list",
        help=(
            "A file with a list of files, one per line, that should be"
            " processed (16kHz mono wav)"
        ),
        nargs="?",
        default=None,
    )
    parser.add_argument(
        "--folder",
        help=("A name of a folder where the files within the list are in."),
        nargs="?",
        default="./",
    )
    parser.add_argument(
        "--outfile",
        help=("A filename to store the results in CSV"),
        nargs="?",
        default=None,
    )
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)

    # create a new experiment
    expr = Experiment(config)
    module = "demo"
    expr.set_module(module)
    util = Util(module)
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    # load the experiment
    expr.load(f"{util.get_save_name()}")
    if args.folder is not None:
        glob_conf.config["DATA"]["test_folder"] = args.folder
    if args.file is None and args.list is None:
        expr.demo(None, False, args.outfile)
    else:
        if args.list is None:
            expr.demo(args.file, False, args.outfile)
        else:
            expr.demo(args.list, True, args.outfile)

    print("DONE")


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # use this if you want to state the config file path on command line



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @Filename : process_database.py --> MESS

import argparse
from pathlib import Path

import pandas as pd

emotion_map = {
    "A": "anger",
    "C": "calm",
    "H": "happiness",
    "S": "sadness",
}

gender_map = {
    "M": "male",
    "F": "female"
}

def main():
    parser = argparse.ArgumentParser(usage="python3 process_database.py database output")
    parser.add_argument("--data_dir", type=str, default="MESS", help="Path to the RAVDESS database")
    parser.add_argument("--out_dir", type=str, default=".", help="Path to the output directory")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    out_dir = Path(args.out_dir)

    paths = list(data_dir.glob("**/*.wav"))
    emotions = [emotion_map[path.stem[0]] for path in paths]
    genders = [gender_map[p.stem[1]] for p in paths]
    speakers = [p.stem[1:3] for p in paths]
    languages = ["english" for p in paths]

    # convert to df
    df = pd.DataFrame({"file": paths, "emotion": emotions, "gender": genders, "speaker": speakers, "language": languages})

    # set speaker M3 as test
    df_test = df[df["speaker"] == "M3"]
    df_train = df.drop(df_test.index)

    # save to csv
    df_train.to_csv(out_dir / "mess_train.csv", index=False)
    df_test.to_csv(out_dir / "mess_test.csv", index=False)
    df.to_csv(out_dir / "mess.csv", index=False)

    print(f"Processed {len(df)} files, {len(df_train)} for training, {len(df_test)} for testing")


if __name__ == "__main__":
    main()



# process_database.ini --> subesco
# modified from https://github.com/Strong-AI-Lab/emotion/blob/master/datasets/SUBESCO/process.py
import argparse
import re
from pathlib import Path
from typing import Dict

import pandas as pd
from sklearn.model_selection import train_test_split

REGEX = re.compile(r"^([MF])_([0-9]+)_([A-Z]+)_S_([0-9]+)_([A-Z]+)_([0-9])$")

emotion_map = {
    "ANGRY": "anger",
    "DISGUST": "disgust",
    "HAPPY": "happiness",
    "NEUTRAL": "neutral",
    "FEAR": "fear",
    "SURPRISE": "surprise",
    "SAD": "sadness",
}

gender_map = {
    'M': "male",
    'F': "female"
}

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="SUBESCO")
    parser.add_argument("--output_dir", type=str, default=".")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    # Load data
    wav_files = list(data_dir.glob("**/*.wav"))
    keys = ["file", "emotion", "speaker", "gender", "language"]
    annot: Dict[str, Dict[str, str]] = {x: {} for x in keys}

    file = [wav_file for wav_file in wav_files if REGEX.match(wav_file.stem) is not None]
    emotion = [emotion_map[REGEX.match(wav_file.stem)[5]] for wav_file in wav_files if REGEX.match(wav_file.stem) is not None]
    speaker_idx = [REGEX.match(wav_file.stem)[2] for wav_file in wav_files if REGEX.match(wav_file.stem) is not None]
    speaker = [REGEX.match(wav_file.stem)[3] for wav_file in wav_files if REGEX.match(wav_file.stem) is not None]
    gender = [gender_map[REGEX.match(wav_file.stem)[1]] for wav_file in wav_files if REGEX.match(wav_file.stem) is not None]
    language = ["bangla" for wav_file in wav_files if REGEX.match(wav_file.stem) is not None]

    # for wav_file in wav_files:
    #     match = REGEX.match(wav_files.stem)
    #     if match is None:
    #         raise ValueError(f"Invalid filename: {wav_file.stem}")
    #     else:
    #         annot["file"][wav_file.stem] = wav_file
    #         annot["emotion"][wav_file.stem] = emotion_map[match[5]]
    #         annot["speaker"][wav_file.stem] = match[2]
    #         annot["gender"][wav_file.stem] = match[1]
    #     # language is same for all files
    #     # annot["language"][wav_file.stem] = "bangla"

    # # collect all data
    # annot = {k: list(v.values()) for k, v in annot.items()} 
    
    # Create dataframe
    df = pd.DataFrame({"file": file, "emotion": emotion, "gender": gender, 
                       "speaker": speaker, "speaker_idx": speaker_idx, 
                       "language": language})
    
    # allocate speaker 9 and 10 from each male and female gender to test set
    test_df = df[df["speaker_idx"].isin(["09", "10"])]
    train_df = df.drop(test_df.index)
    train_df.to_csv('subesco_train.csv', index=False)
    test_df.to_csv('subesco_test.csv', index=False)

    print(f"Total: {len(df)}, Train: {len(train_df)}, Test: {len(test_df)}")
    
if __name__ == "__main__":
    main()
